import numpy as np
import torch
import os
import traceback
import time
import nrrd
import sys
import matplotlib.pyplot as plt
import logging
import argparse
import torch.nn.functional as F
from scipy.stats import norm
from torch.utils.data import DataLoader
from tqdm import tqdm
import termcolor
from torch.autograd import Variable
from torch.nn.parallel.data_parallel import data_parallel
from scipy.ndimage.measurements import label
from scipy.ndimage import center_of_mass
from net.sanet import SANet
from dataset.collate import train_collate, test_collate, eval_collate
# from dataset.bbox_reader import BboxReader
from dataset.bbox_reader_UCLA import BboxReader
from config import config
import pandas as pd
import pickle
from evaluationScript.noduleCADEvaluationLUNA16 import noduleCADEvaluation
from utils.prepare_dataset_UCLA import seperate_cnds


this_module = sys.modules[__name__]
os.environ['CUDA_VISIBLE_DEVICES'] = '2'


parser = argparse.ArgumentParser()
parser.add_argument('--net', '-m', metavar='NET', default=config['net'],
                    help='neural net')
parser.add_argument("--mode", type=str, default = 'eval',
                    help="you want to test or val")
parser.add_argument("--weight", type=str, default='./model.ckpt',
                    help="path to model weights to be used")
parser.add_argument("--dicom-path", type=str, default=None,
                    help="path to dicom files of patient")
parser.add_argument("--out-dir", type=str, default=config['out_dir'],
                    help="path to save the results")
parser.add_argument("--test-set-name", type=str, default=config['test_set_name'],
                    help="path to save the results")


def main(cur_model_data_dir, out_dir, cur_cnd):
    args = parser.parse_args()
    logging.basicConfig(format='[%(levelname)s][%(asctime)s] %(message)s', level=logging.INFO)
    if args.mode == 'eval':
        print('Running in eval mode')
        print('Reading data from:', cur_model_data_dir)
        print('Test set name:', test_set_name)
        print('output dir:', out_dir)

        num_workers = 1
        initial_checkpoint = args.weight
        print('Initial checkpoint:', initial_checkpoint)
        net = args.net
        print('Network args:', net)        
        
        net = getattr(this_module, net)(config)
        net = net.cuda()
        termcolor.cprint('Network Instantiated | on CUDA', 'green')
        if initial_checkpoint:
            print('[Loading model from %s]' % initial_checkpoint)
            checkpoint = torch.load(initial_checkpoint)
            epoch = checkpoint['epoch']
            net.load_state_dict(checkpoint['state_dict'])
            print('Model weights loaded!')
        else:
            print('No model weight file specified. Exiting...')
            return None

        save_dir = os.path.join(out_dir, 'res', str(epoch))
        print('Saving dir:', save_dir)
        if not os.path.exists(save_dir):
            os.makedirs(save_dir)
        if not os.path.exists(os.path.join(save_dir, 'FROC')):
            os.makedirs(os.path.join(save_dir, 'FROC'))
    
        dataset = BboxReader(cur_model_data_dir, test_set_name, config, mode='eval', segment_lung=USE_MASK, add_cnd=cur_cnd)
        print('Dataset created')
        test_loader = DataLoader(dataset, batch_size=1, shuffle=False,
                                 num_workers=num_workers, pin_memory=False, collate_fn=train_collate)
        # for i, (input, truth_bboxes, truth_labels) in enumerate(dataset):
        #     print(type(input), input.min(), input.max())
        #     break
        return eval(net, test_loader, save_dir)
    else:
        logging.error('Mode %s is not supported' % (args.mode))


def eval(net, dataset, save_dir=None):
    net.set_mode('eval')
    net.use_rcnn = True

    print('Total # of eval data %d' % (len(dataset)))
    for i, (input, truth_bboxes, truth_labels) in enumerate(dataset):
        try:
            input = Variable(input).cuda()
            print('input on cuda size:', input.shape)
            truth_bboxes = np.array(truth_bboxes)
            truth_labels = np.array(truth_labels)
            pid = dataset.dataset.filenames[i]

            print('[%d] Predicting %s' % (i, pid))

            with torch.no_grad():
                net.forward(input, truth_bboxes, truth_labels)

            detections = net.rpn_proposals.cpu().numpy()
            print('final detections shape:', detections.shape)

            if len(detections):
                detections = detections[:, 1:-1]
                np.save(os.path.join(save_dir, '%s_detections.npy' % (pid)), detections)

            # clear gpu memory
            del input, truth_bboxes, truth_labels
            torch.cuda.empty_cache()

        except Exception as e:
            del input, truth_bboxes, truth_labels
            torch.cuda.empty_cache()
            traceback.print_exc()
            return
   
    print('All detections done!')
    # Generate prediction csv for the use of performning FROC analysis
    res = []
    for pid in dataset.dataset.filenames:
        if os.path.exists(os.path.join(save_dir, '%s_detections.npy' % (pid))):
            detections = np.load(os.path.join(save_dir, '%s_detections.npy' % (pid)))
            detections = detections[:, [3, 2, 1, 4, 0]]
            names = np.array([[pid]] * len(detections))
            res.append(np.concatenate([names, detections], axis=1))
    
    res = np.concatenate(res, axis=0)
    col_names = ['pid','center_x','center_y','center_z','diameter', 'probability']
    eval_dir = os.path.join(save_dir, 'FROC')
    res_path = os.path.join(eval_dir, 'results.csv')
    
    df = pd.DataFrame(res, columns=col_names)
    df.to_csv(res_path, index=False)
    # start evaluating
    if not os.path.exists(os.path.join(eval_dir, 'res')):
        os.makedirs(os.path.join(eval_dir, 'res'))

    annotations_filename = '/workspace/SANet/data/ucla_filtered_testUID_70_30_split_annotGT.csv'
    val_path = '/workspace/SANet/data/ucla_filtered_testUID_70_30_split_pids.txt'
    return noduleCADEvaluation(annotations_filename, res_path, val_path, os.path.join(eval_dir, 'res'), return_dict=True)


if __name__ == '__main__':
    USE_MASK = '/datasets/bodymask/k2_d100_st1.0_lungonly'
    out_dir = './results_normCnd_pred-in-DiamBox'
    test_UIDS = '/datasets/uids/70-30_split/ucla_testUID_70_30_split.txt' # to generate cnds in dataset
    cnds = seperate_cnds(test_UIDS)
    assert len(cnds['k1_d10_st1']) == len(cnds['k1_d25_st1']) == len(cnds['k1_d100_st1'])
    test_set_name = './data/ucla_filtered_testUID_70_30_split_pids.txt' # uids to read from overall test set

    """
    # build path to dataset
    data_dir = {'Unnorm':'/datasets/data_st1.0_merged'}
    norm_cases, excludeCnd = '/norm_data', ['UnNorm']
    for case in os.listdir(norm_cases):
        if case not in excludeCnd:
            model_cls = os.listdir(os.path.join(norm_cases, case))
            for model in model_cls:
                model_name = model.split('-70-30-')[0].split('test_')[-1]
                if (model_name not in data_dir) and (model_name not in excludeCnd):
                    path_to_cases =  os.path.join(norm_cases, case, model, 'merged_cases', 'test_st1.0_norm')
                    assert len(os.listdir(path_to_cases)) == 240, "Number of cases not equal to 240!!"
                    data_dir[model_name] = path_to_cases
    """

    data_dir = {'SNGAN_withCBAM_cand_Attn':'/norm_data/GANs/test_SNGAN_withCandidateAtten-CBAM_70-30-Split_tileStitch-32x64x64_Ref-k2d100/merged_cases/test_st1.0_norm'}
    
    # store resulting output
    if os.path.exists(os.path.join(out_dir, 'FROC_dict_LT.pkl')):
        print('Storing in already existing dictionary!')
        master_dict = pickle.load(open(os.path.join(out_dir, 'FROC_dict_LT.pkl'), "rb"))
    else:
        master_dict = {}

    # run inference
    for model in data_dir:
        print('Processing model:', model)
        master_dict[model] = {}
        for cnd in cnds:
            print('Running condition:', cnd)
            cnd_dir = os.path.join(out_dir, model, cnd)
            if not os.path.exists(cnd_dir):
                os.makedirs(cnd_dir)
            out_result = main(cur_model_data_dir=data_dir[model], out_dir=cnd_dir, cur_cnd=cnd)
            master_dict[model][cnd] = out_result

    with open(os.path.join(out_dir, 'FROC_dict_LT_v1.pkl'), 'wb') as file:
        pickle.dump(master_dict, file)

    """
    for cnd in cnds:
        print('Running condition:', cnd)
        cnd_dir = os.path.join(out_dir, cnd)
        if not os.path.exists(cnd_dir):
            os.makedirs(cnd_dir)
        out_result = main(out_dir=cnd_dir, cur_cnd=cnd)        
        master_dict[cnd] = out_result
    with open(os.path.join(out_dir, 'FROC_dict.pkl'), 'wb') as file:
        pickle.dump(master_dict, file)
    """
    # python test_normCnd.py --weight='./model.ckpt'



import os
import os.path as osp
import logging
from collections import OrderedDict
import json
import lung_analysis_pipeline.utils.util as util

def parse(opt_path, is_train=False):
    # remove comments starting with '//'
    json_str = ''
    with open(opt_path, 'r') as f:
        for line in f:
            line = line.split('//')[0] + '\n'
            json_str += line
    opt = json.loads(json_str, object_pairs_hook=OrderedDict)

    check_order(opt)

    # results_root = opt['output']['output_location']
    # opt['output']['log'] = results_root
    # util.mkdir(results_root) 

    # For feature extraction, redirect to another file if the output path already exists 
    # for operation in ['feature_extraction', 'detection']:
    if 'feature_extraction' in opt: 
        for extractor in ['deep_feature', 'radiomics']:
            if extractor in opt['feature_extraction']:  
                opt['feature_extraction'][extractor]['output_path'] and os.path.exists(opt['feature_extraction'][extractor]['output_path'])
                filename = os.path.basename(opt['feature_extraction'][extractor]['output_path'])

                # Extract the file name without extension and append timestamp 
                file_name_without_ext = os.path.splitext(filename)[0]
                new_file_name = file_name_without_ext + util.get_timestamp() + os.path.splitext(filename)[1]

                # Get the directory of the path and new path 
                directory = os.path.dirname(opt['feature_extraction'][extractor]['output_path'])
                opt['feature_extraction'][extractor]['output_path'] = os.path.join(directory, new_file_name)
                
                logger = logging.getLogger('base')
                logger.info('File already exists. Rename it to [{:s}]'.format(opt['feature_extraction'][extractor]['output_path']))
   
    if 'detection' in opt and 'output_path' in opt['detection'] and os.path.exists(opt['detection']['output_path']): 
        filename = os.path.basename(opt['detection']['output_path'])

        # Extract the file name without extension and append timestamp 
        file_name_without_ext = os.path.splitext(filename)[0]
        new_file_name = file_name_without_ext + util.get_timestamp() + os.path.splitext(filename)[1]

        # Get the directory of the path and new path 
        directory = os.path.dirname(opt['detection']['output_path'])
        opt['detection']['output_path'] = os.path.join(directory, new_file_name)
        
        logger = logging.getLogger('base')
        logger.info('File already exists. Rename it to [{:s}]'.format(opt['detection']['output_path']))
    
    gpu_list = ','.join(str(x) for x in opt['gpu_ids']) if opt['gpu_ids'] else ""
    os.environ['CUDA_VISIBLE_DEVICES'] = gpu_list
    # print('export CUDA_VISIBLE_DEVICES=' + gpu_list)
    
    
    return opt


"""
Check if specified operations are allowed. 
"""
def check_order(opt): 
    order = opt['order']
    for i, ops in enumerate(order): 
        if ops not in opt:
            raise ValueError('Please specify {} operation.'.format(ops))
        if ops == 'normalization': 
            if i != 0: 
                raise ValueError('Normalization operation order not supported!')
            if opt[ops]['model'] == 'bm3d' or "weights" not in opt[ops]: 
                raise ValueError('Need to specify path to normalization model weights.')
        elif ops == 'detection': 
            if "weights" not in opt[ops]: 
                raise ValueError('Need to specify path to detection model weights.')
        elif ops == 'segmentation': 
            # there has to exist detection before seg or use detection is set to false 
            if opt[ops]['use_decetion_bbox'] and not 'detection' in order[:i]: 
                raise ValueError('Segmentation operation order not supported!')
            if 'model' not in opt[ops]: 
                raise ValueError("Need to specify a segmentation model.")
            # check if the specified model is supported 
            if opt[ops]['model'] != 'unet': 
                raise NotImplementedError('Model [{:s}] not implemented.'.format(opt[ops]['model']))
            if 'weights' not in opt[ops]: 
                raise ValueError('Need to specify path to segmentation model weights.')
        elif ops == 'feature_extraction': 
            if 'roi_source' not in opt[ops]:
                raise ValueError('Segmentation operation order not supported! Need to specify source of ROI.')
            if opt[ops]['roi_source'] == 'segmentation' and not 'segmentation' in order[:i]: 
                raise ValueError('Feature extraction order not supported. Need to specify a segmentation method.')
            if opt[ops]['roi_source'] == 'detection' and not 'detection' in order[:i]: 
                raise ValueError('Feature extraction order not supported. Need to specify a detection method.')
            if opt[ops]['roi_source'] == 'user' and 'feature_roi_centroid' not in opt['dataset']: 
                raise ValueError('Feature extraction order not supported. Need to specify a ROI location.')
            if opt[ops]['roi_source'] == 'user' and 'feature_mask' not in opt['dataset']: 
                raise ValueError('Feature extraction order not supported. Need to specify a mask location.')
            if opt[ops]['deep_feature']: 
                if 'model' not in opt[ops]['deep_feature']: 
                    raise ValueError('Need to specify a deep learning model for feature extraction.')
                if opt[ops]['deep_feature']['model'] != 'hscnn': 
                    raise ValueError('Deep feature extraction model not supported.')
                if 'weights' not in opt[ops]['deep_feature']: 
                    raise ValueError('Need to specify path to deep feature extraction model weights.')


import torch 
import pkg_resources

from .net.sanet import SANet 
from .config import config
import logging
logger = logging.getLogger('base')


def create_detect_model(opt): 
    # define config 

    # load pre-trained model 
    # path_to_weight = 'model.ckpt'
    # path_to_weight = pkg_resources.resource_filename(__name__, path_to_weight)
    path_to_weight = opt['detection']['model']
    
    checkpoint = torch.load(path_to_weight)
    model = SANet(config)
    model.cuda()
    model.load_state_dict(checkpoint['state_dict'])
    logger.info('Nodule detection model {:s} is created.'.format(model.__class__.__name__)) 

    return model 



import torch
from torch import nn
from models import resnet


def generate_model(opt):
    assert opt.model in [
        'resnet'
    ]

    if opt.model == 'resnet':
        assert opt.model_depth in [10, 18, 34, 50, 101, 152, 200]
        
        if opt.model_depth == 10:
            model = resnet.resnet10(
                sample_input_W=opt.input_W,
                sample_input_H=opt.input_H,
                sample_input_D=opt.input_D,
                shortcut_type=opt.resnet_shortcut,
                no_cuda=opt.no_cuda,
                num_seg_classes=opt.n_seg_classes)
        elif opt.model_depth == 18:
            model = resnet.resnet18(
                sample_input_W=opt.input_W,
                sample_input_H=opt.input_H,
                sample_input_D=opt.input_D,
                shortcut_type=opt.resnet_shortcut,
                no_cuda=opt.no_cuda,
                num_seg_classes=opt.n_seg_classes)
        elif opt.model_depth == 34:
            model = resnet.resnet34(
                sample_input_W=opt.input_W,
                sample_input_H=opt.input_H,
                sample_input_D=opt.input_D,
                shortcut_type=opt.resnet_shortcut,
                no_cuda=opt.no_cuda,
                num_seg_classes=opt.n_seg_classes)
        elif opt.model_depth == 50:
            model = resnet.resnet50(
                sample_input_W=opt.input_W,
                sample_input_H=opt.input_H,
                sample_input_D=opt.input_D,
                shortcut_type=opt.resnet_shortcut,
                no_cuda=opt.no_cuda,
                num_seg_classes=opt.n_seg_classes)
        elif opt.model_depth == 101:
            model = resnet.resnet101(
                sample_input_W=opt.input_W,
                sample_input_H=opt.input_H,
                sample_input_D=opt.input_D,
                shortcut_type=opt.resnet_shortcut,
                no_cuda=opt.no_cuda,
                num_seg_classes=opt.n_seg_classes)
        elif opt.model_depth == 152:
            model = resnet.resnet152(
                sample_input_W=opt.input_W,
                sample_input_H=opt.input_H,
                sample_input_D=opt.input_D,
                shortcut_type=opt.resnet_shortcut,
                no_cuda=opt.no_cuda,
                num_seg_classes=opt.n_seg_classes)
        elif opt.model_depth == 200:
            model = resnet.resnet200(
                sample_input_W=opt.input_W,
                sample_input_H=opt.input_H,
                sample_input_D=opt.input_D,
                shortcut_type=opt.resnet_shortcut,
                no_cuda=opt.no_cuda,
                num_seg_classes=opt.n_seg_classes)
    
    if not opt.no_cuda:
        if len(opt.gpu_id) > 1:
            model = model.cuda() 
            model = nn.DataParallel(model, device_ids=opt.gpu_id)
            net_dict = model.state_dict() 
        else:
            import os
            os.environ["CUDA_VISIBLE_DEVICES"]=str(opt.gpu_id[0])
            model = model.cuda() 
            model = nn.DataParallel(model, device_ids=None)
            net_dict = model.state_dict()
    else:
        net_dict = model.state_dict()
    
    """
    # load pretrain
    if opt.phase != 'test' and opt.pretrain_path:
        print ('loading pretrained model {}'.format(opt.pretrain_path))
        pretrain = torch.load(opt.pretrain_path)
        # get encoder part state dict from pretraining
        pretrain_dict = {k: v for k, v in pretrain['state_dict'].items() if k in net_dict.keys()}
        # update new model state dict with old state dict weights and biases
        net_dict.update(pretrain_dict)
        # load update state dic
        model.load_state_dict(net_dict)
        new_parameters = []

        for pname, p in model.named_parameters():
            for layer_name in opt.new_layer_names:
                if pname.find(layer_name) >= 0: # will give value -1 if new layer doesnot exist, else give some other value
                    new_parameters.append(p)
                    break

        new_parameters_id = list(map(id, new_parameters))
        base_parameters = list(filter(lambda p: id(p) not in new_parameters_id, model.parameters()))
        parameters = {'base_parameters': base_parameters, 
                      'new_parameters': new_parameters}
        return model, parameters
    """ 
    return model, model.parameters()



import math
import torch
import torch.nn as nn
from termcolor import cprint
from collections import OrderedDict


class SpatialCGNL(nn.Module):
    """Spatial CGNL block with dot production kernel for image classfication.
    """
    def __init__(self, inplanes, planes, use_scale=False, groups=None):
        self.use_scale = use_scale
        self.groups = groups

        super(SpatialCGNL, self).__init__()
        # conv theta
        self.t = nn.Conv3d(inplanes, planes, kernel_size=1, stride=1, bias=False)
        # conv phi
        self.p = nn.Conv3d(inplanes, planes, kernel_size=1, stride=1, bias=False)
        # conv g
        self.g = nn.Conv3d(inplanes, planes, kernel_size=1, stride=1, bias=False)
        # conv z
        self.z = nn.Conv3d(planes, inplanes, kernel_size=1, stride=1,
                                                  groups=self.groups, bias=False)
        self.gn = nn.GroupNorm(num_groups=self.groups, num_channels=inplanes)

        if self.use_scale:
            cprint("=> WARN: SpatialCGNL block uses 'SCALE'", \
                   'yellow')
        if self.groups:
            cprint("=> WARN: SpatialCGNL block uses '{}' groups".format(self.groups), \
                   'yellow')

    def kernel(self, t, p, g, b, c, zs, h, w):
        """The linear kernel (dot production).

        Args:
            t: output of conv theata
            p: output of conv phi
            g: output of conv g
            b: batch size
            c: channels number
            h: height of featuremaps
            w: width of featuremaps
        """
        t = t.contiguous().view(b, 1, c * zs * h * w)
        p = p.contiguous().view(b, 1, c * zs * h * w)
        g = g.contiguous().view(b, c * zs * h * w, 1)

        att = torch.bmm(p, g)

        if self.use_scale:
            att = att.div((c*zs*h*w)**0.5)

        x = torch.bmm(att, t)
        x = x.view(b, c, zs, h, w)

        return x

    def forward(self, x):
        residual = x

        t = self.t(x)
        p = self.p(x)
        g = self.g(x)

        b, c, zs, h, w = t.size()

        if self.groups and self.groups > 1:
            _c = int(zs / self.groups)

            ts = torch.split(t, split_size_or_sections=_c, dim=2)
            ps = torch.split(p, split_size_or_sections=_c, dim=2)
            gs = torch.split(g, split_size_or_sections=_c, dim=2)

            _t_sequences = []
            for i in range(self.groups):
                _x = self.kernel(ts[i], ps[i], gs[i],
                                 b, c, _c, h, w)
                _t_sequences.append(_x)

            x = torch.cat(_t_sequences, dim=2)
        else:
            x = self.kernel(t, p, g,
                            b, c, zs, h, w)

        x = self.z(x)
        x = self.gn(x) + residual

        return x


class SpatialCGNLx(nn.Module):
    """Spatial CGNL block with Gaussian RBF kernel for image classification.
    """
    def __init__(self, inplanes, planes, use_scale=False, groups=None, order=2):
        self.use_scale = use_scale
        self.groups = groups
        self.order = order

        super(SpatialCGNLx, self).__init__()
        # conv theta
        self.t = nn.Conv3d(inplanes, planes, kernel_size=1, stride=1, bias=False)
        # conv phi
        self.p = nn.Conv3d(inplanes, planes, kernel_size=1, stride=1, bias=False)
        # conv g
        self.g = nn.Conv3d(inplanes, planes, kernel_size=1, stride=1, bias=False)
        # conv z
        self.z = nn.Conv3d(planes, inplanes, kernel_size=1, stride=1,
                                                  groups=self.groups, bias=False)
        self.gn = nn.GroupNorm(num_groups=self.groups, num_channels=inplanes)

        if self.use_scale:
            cprint("=> WARN: SpatialCGNLx block uses 'SCALE'", \
                   'yellow')
        if self.groups:
            cprint("=> WARN: SpatialCGNLx block uses '{}' groups".format(self.groups), \
                   'yellow')

        cprint('=> WARN: The Taylor expansion order in SpatialCGNLx block is {}'.format(self.order), \
               'yellow')

    def kernel(self, t, p, g, b, c, zs, h, w):
        """The non-linear kernel (Gaussian RBF).

        Args:
            t: output of conv theata
            p: output of conv phi
            g: output of conv g
            b: batch size
            c: channels number
            h: height of featuremaps
            w: width of featuremaps
        """

        t = t.view(b, 1, c * zs * h * w)
        p = p.view(b, 1, c * zs * h * w)
        g = g.view(b, c * zs * h * w, 1)

        # gamma
        gamma = torch.Tensor(1).fill_(1e-4)

        # NOTE:
        # We want to keep the high-order feature spaces in Taylor expansion to
        # rich the feature representation, so the l2 norm is not used here.
        #
        # Under the above precondition, the β should be calculated
        # by β = exp(−γ(∥θ∥^2 +∥φ∥^2)).
        # But in the experiments, we found training becomes very difficult.
        # So we simplify the implementation to
        # ease the gradient computation through calculating the β = exp(−2γ).

        # beta
        beta = torch.exp(-2 * gamma)

        t_taylor = []
        p_taylor = []
        for order in range(self.order+1):
            # alpha
            alpha = torch.mul(
                    torch.div(
                    torch.pow(
                        (2 * gamma),
                        order),
                        math.factorial(order)),
                        beta)

            alpha = torch.sqrt(
                        alpha.cuda())

            _t = t.pow(order).mul(alpha)
            _p = p.pow(order).mul(alpha)

            t_taylor.append(_t)
            p_taylor.append(_p)

        t_taylor = torch.cat(t_taylor, dim=1)
        p_taylor = torch.cat(p_taylor, dim=1)

        att = torch.bmm(p_taylor, g)

        if self.use_scale:
            att = att.div((c*zs*h*w)**0.5)

        att = att.view(b, 1, int(self.order+1))
        x = torch.bmm(att, t_taylor)
        x = x.view(b, c, zs, h, w)

        return x

    def forward(self, x):
        residual = x

        t = self.t(x)
        p = self.p(x)
        g = self.g(x)

        b, c, zs, h, w = t.size()

        if self.groups and self.groups > 1:
            _c = int(c / self.groups)

            ts = torch.split(t, split_size_or_sections=_c, dim=1)
            ps = torch.split(p, split_size_or_sections=_c, dim=1)
            gs = torch.split(g, split_size_or_sections=_c, dim=1)

            _t_sequences = []
            for i in range(self.groups):
                _x = self.kernel(ts[i], ps[i], gs[i],
                                 b, _c, zs, h, w)
                _t_sequences.append(_x)

            x = torch.cat(_t_sequences, dim=1)
        else:
            x = self.kernel(t, p, g,
                            b, c, zs, h, w)

        x = self.z(x)
        x = self.gn(x) + residual

        return x


class SpatialNL(nn.Module):
    """Spatial NL block for image classification.
       [https://github.com/facebookresearch/video-nonlocal-net].
    """
    def __init__(self, inplanes, planes, use_scale=False):
        self.use_scale = use_scale

        super(SpatialNL, self).__init__()
        self.t = nn.Conv2d(inplanes, planes, kernel_size=1, stride=1, bias=False)
        self.p = nn.Conv2d(inplanes, planes, kernel_size=1, stride=1, bias=False)
        self.g = nn.Conv2d(inplanes, planes, kernel_size=1, stride=1, bias=False)
        self.softmax = nn.Softmax(dim=2)
        self.z = nn.Conv2d(planes, inplanes, kernel_size=1, stride=1, bias=False)
        self.bn = nn.BatchNorm2d(inplanes)

        if self.use_scale:
            cprint("=> WARN: SpatialNL block uses 'SCALE' before softmax", 'yellow')

    def forward(self, x):
        residual = x

        t = self.t(x)
        p = self.p(x)
        g = self.g(x)

        b, c, h, w = t.size()

        t = t.view(b, c, -1).permute(0, 2, 1)
        p = p.view(b, c, -1)
        g = g.view(b, c, -1).permute(0, 2, 1)

        att = torch.bmm(t, p)

        if self.use_scale:
            att = att.div(c**0.5)

        att = self.softmax(att)
        x = torch.bmm(att, g)

        x = x.permute(0, 2, 1)
        x = x.contiguous()
        x = x.view(b, c, h, w)

        x = self.z(x)
        x = self.bn(x) + residual

        return x


import numpy as np
from .util import box_transform, box_transform_inv, clip_boxes
import itertools
import torch.nn.functional as F
import torch
from torch.autograd import Variable
try:
    from lung_analysis_pipeline.NoduleDetect_SANet.utils.pybox import *
except ImportError:
    print('Warning: C++ module import failed! This should only happen in deployment')
    from lung_analysis_pipeline.NoduleDetect_SANet.utils.util import py_nms as torch_nms
    from lung_analysis_pipeline.NoduleDetect_SANet.utils.util import py_box_overlap as torch_overlap

def make_rpn_windows(f, cfg):
    """
    Generating anchor boxes at each voxel on the feature map,
    the center of the anchor box on each voxel corresponds to center
    on the original input image.

    return
    windows: list of anchor boxes, [z, y, x, d, h, w]
    """
    stride = cfg['stride']
    anchors = np.asarray(cfg['anchors'])
    offset = (float(stride) - 1) / 2
    _, _, D, H, W = f.shape
    oz = np.arange(offset, offset + stride * (D - 1) + 1, stride)
    oh = np.arange(offset, offset + stride * (H - 1) + 1, stride)
    ow = np.arange(offset, offset + stride * (W - 1) + 1, stride)

    windows = []
    for z, y , x , a in itertools.product(oz, oh , ow , anchors):
        windows.append([z, y , x , a[0], a[1], a[2]])
    windows = np.array(windows)

    return windows

def rpn_nms(cfg, mode, inputs, window, logits_flat, deltas_flat):
    if mode in ['train',]:
        nms_pre_score_threshold = cfg['rpn_train_nms_pre_score_threshold']
        nms_overlap_threshold   = cfg['rpn_train_nms_overlap_threshold']

    elif mode in ['eval', 'valid', 'test',]:
        nms_pre_score_threshold = cfg['rpn_test_nms_pre_score_threshold']
        nms_overlap_threshold   = cfg['rpn_test_nms_overlap_threshold']

    else:
        raise ValueError('rpn_nms(): invalid mode = %s?'%mode)

    logits = torch.sigmoid(logits_flat).data.cpu().numpy()
    deltas = deltas_flat.data.cpu().numpy()
    batch_size, _, depth, height, width = inputs.size()

    proposals = []
    for b in range(batch_size):
        proposal = [np.empty((0, 8),np.float32),]
        ps = logits[b, : , 0].reshape(-1, 1)
        ds = deltas[b, :, :]
        # Only those anchor boxes larger than a pre-defined threshold
        # will be chosen for nms computation
        index = np.where(ps[:, 0] > nms_pre_score_threshold)[0]
        if len(index) > 0:
            p = ps[index]
            d = ds[index]
            w = window[index]
            box = rpn_decode(w, d, cfg['box_reg_weight'])
            box = clip_boxes(box, inputs.shape[2:])
            output = np.concatenate((p, box),1)
            output = torch.from_numpy(output)
            output, keep = torch_nms(output, nms_overlap_threshold)
            prop = np.zeros((len(output), 8),np.float32)
            prop[:, 0] = b
            prop[:, 1:8] = output
            proposal.append(prop)

        proposal = np.vstack(proposal)
        proposals.append(proposal)

    proposals = np.vstack(proposals)

    # Just in case if there is no proposal, we still return a Tensor,
    # torch.from_numpy() cannot take input with 0 dim
    if len(proposals) != 0:
        proposals = Variable(torch.from_numpy(proposals)).cuda()
        return proposals
    else:
        return Variable(torch.rand([0, 8])).cuda()

    return proposals

def rpn_encode(window, truth_box, weight):
    return box_transform(window, truth_box, weight)

def rpn_decode(window, delta, weight):
    return box_transform_inv(window, delta, weight)



from torch.utils.data import DataLoader, random_split
import torch.optim as optim
import matplotlib.pyplot as plt
from dataloader import HscnnDataset
import pickle
import os
import torch
import pandas as pd
import logging
import torch.nn as nn
import numpy as np
from model import HSCNN
from keras.utils import np_utils # Used only for converting true vectors to one-hot for AUC calculations
from itertools import cycle
import torch.nn.functional as F
from sklearn.metrics import confusion_matrix, precision_score, accuracy_score, recall_score, \
    roc_curve, roc_auc_score, auc
from scipy import interp
# Transforms
from torchio.transforms import (
    RandomFlip,
    RandomAffine,
    RandomMotion,
    RescaleIntensity,
    Compose,
)

LOG_FILENAME = 'HSCNN_LOG_FOLDS_01_23_2023.log'
logging.basicConfig(filename=LOG_FILENAME, level=logging.INFO,
                    format='%(levelname)s: %(message)s')


class TrainModel:

    def __init__(self, input_channel=1, num_low_level_tasks=3, low_level_outputs=[2, 2, 2], loss_weights=[0.1, 0.1, 0.1], malignancy_class=2, num_epochs=300, batch_size=6, val_per=0.2):
        self.in_channel = input_channel
        self.num_low_level_tasks = num_low_level_tasks
        self.low_level_outputs = low_level_outputs
        self.malignancy_class = malignancy_class
        self.num_epochs = num_epochs
        self.batch_size = batch_size
        self.loss_weights = loss_weights
        self.val_per = val_per
        self.device = self._get_device()
        self.thresh = 0.5

    def _get_device(self):
        train_on_gpu = torch.cuda.is_available()
        if not train_on_gpu:
            device = torch.device("cpu")
            logging.info("Running on {}".format(device))
        else:
            device = torch.device("cuda:0")
            logging.info("Running on {}".format(torch.cuda.get_device_name(0)))
        return device
	
    # make sure input data is clipped in range [-1000, 500] HU
    def _get_default_transforms(self):
        io_transforms = Compose([
            RescaleIntensity((0, 1)),  # so that there are no negative values for RandomMotion
            # RandomMotion(),
            RandomFlip(axes=(1,)),
            # RandomAffine(scales=(0.9, 1.2), degrees=(10), isotropic=False, default_pad_value='otsu', image_interpolation='bspline')
        ])
        return io_transforms

    def _weights_init(self, model):
        classname = model.__class__.__name__
        if classname.find('Conv3d') != -1:
            # model.weight.data.normal_(0.0, 0.02)
            nn.init.xavier_uniform_(model.weight,  gain=nn.init.calculate_gain('relu'))
            model.bias.data.fill_(0)
        # elif classname.find('BatchNorm3d') != -1:
            # model.weight.data.normal_(1.0, 0.02)
            # model.bias.data.fill_(0)

    def _get_HSCNN(self, in_channel, num_low_level_tasks, low_level_outputs, malignancy_class, initialize_weights=True):
        # Get the Model
        model = HSCNN(in_channel, num_low_level_tasks, low_level_outputs, malignancy_class).to(self.device)
        if initialize_weights:
            model.apply(self._weights_init)
        return model

    """
    Calculate class weights over dataset
    """
    def _class_weights(self, path_to_label, pos_weight=False):
        data = pd.read_csv(path_to_label)
        weights = {}
        for i in range(self.num_low_level_tasks + 1):
            lbl = data.iloc[:, i].value_counts()
            if pos_weight:
                pos_sample_ratio = lbl[0]/lbl[1]
                weights['task_{}_weights'.format(i + 1)] = [pos_sample_ratio]
            else:
                weight_0 = lbl[1]/(lbl[0]+lbl[1])
                weight_1 = lbl[0] / (lbl[0] + lbl[1])
                weights['task_{}_weights'.format(i+1)] = [weight_0, weight_1]
        return weights

    def _get_evaluation_metric(self, y_truth_dict, y_predicted_dict, scores_dict, each_class=False):
        label = ["Sphericity", "Texture", "Margin", "Subtlety", "Calcification", "Malignancy"]
        class_outputs = self.low_level_outputs.copy()
        class_outputs.append(self.malignancy_class)
        evaluations = {}
        for i in range(self.num_low_level_tasks+1):
            y_true = y_truth_dict['label_{}'.format(i+1)]
            one_hot_true = np_utils.to_categorical(y_true, num_classes=class_outputs[i])
            y_pred = y_predicted_dict['label_{}'.format(i+1)]
            score = np.array(scores_dict['label_{}'.format(i+1)])
            # Calculate the metrics
            t_n, f_p, f_n, t_p = confusion_matrix(y_true, y_pred).ravel()
            recall_sensitivity = recall_score(y_true, y_pred)
            #recall_sensitivity = t_p / (t_p + f_n)
            specificity = t_n / (t_n + f_p)
            accuracy = accuracy_score(y_true, y_pred)
            precision = precision_score(y_true, y_pred)
            #precision = t_p / (t_p + f_p)
            #f1_score = (2 * precision * recall_sensitivity) / (precision + recall_sensitivity)
            model_auc = roc_auc_score(one_hot_true, score, average='weighted')
            evaluations[label[i]] = [accuracy, precision, recall_sensitivity, specificity, model_auc]
        return evaluations

    def _start_model(self, model, optimizer, criterions, scheduler, lbl_weights, loss_weights, train_loader=None, val_loader=None, test_loader=None, fold=0):
        if train_loader:
            train_losses, val_losses, accuracy_list, specificity_list, recall_list, f1_score_list = [], [], [], [], [], []
            valid_loss_min = np.Inf
            # Start with inital weights
            best_model_wts = False
            best_epch_at = 0
            aucs_across_folds = []
            for epoch in range(self.num_epochs):
                # Keep track of lossess
                running_train_loss, running_val_loss = 0.0, 0.0
                epoch_loss = []
                # Training loop
                for index, data in enumerate(train_loader):
                    images, labels = data['img'].to(self.device), data['label']
                    per_label_for_each_image = [labels[:, i].to(self.device, dtype=torch.long) for i in range(self.num_low_level_tasks + 1)]
                    optimizer.zero_grad()
                    output = model(images)
                    running_loss_subtask = 0
                    # Calculate loss for each sub-network
                    for i in range(self.num_low_level_tasks+1):
                        criterion = criterions['criterion_{}'.format(i+1)]
                        running_loss_subtask += criterion(output[i], per_label_for_each_image[i]) * loss_weights[i]

                    loss = running_loss_subtask / (self.num_low_level_tasks+1)
                    loss.backward()
                    optimizer.step()
                    epoch_loss.append(float(loss.item() * images.size(0)))
                    running_train_loss += float(loss.item()) * images.size(0)

                scheduler.step(np.mean(epoch_loss))
                # Validation loop
                with torch.no_grad():
                    model.eval()
                    # Store the predictions and scores
                    predicted_labels_dict = {'label_1': [], 'label_2': [], 'label_3': [], 'label_4': [], 'label_5': [], 'label_6': []}
                    true_labels_dict = {'label_1': [], 'label_2': [], 'label_3': [], 'label_4': [], 'label_5': [], 'label_6': []}
                    scores_dict = {'label_1': [], 'label_2': [], 'label_3': [], 'label_4': [], 'label_5': [], 'label_6': []}

                    for index, data in enumerate(val_loader):
                        images, labels = data['img'].to(self.device), data['label']
                        per_label_for_each_image = [labels[:, i].to(self.device, dtype=torch.long) for i in range(self.num_low_level_tasks + 1)]
                        output = model(images)
                        val_loss_subtask = 0
                        for i in range(self.num_low_level_tasks+1):
                            criterion = criterions['criterion_{}'.format(i+1)]
                            val_loss_subtask += criterion(output[i], per_label_for_each_image[i]) * loss_weights[i]
                            per_task_out = F.softmax(output[i].cpu(), dim=1)
                            top_ps, top_class = per_task_out.topk(1, dim=1)
                            predicted_labels_dict['label_{}'.format(i + 1)].extend(list(top_class.flatten().numpy()))
                            true_labels_dict['label_{}'.format(i + 1)].extend(list(per_label_for_each_image[i].cpu().numpy()))
                            scores_dict['label_{}'.format(i + 1)].extend(per_task_out.numpy().tolist())
                        loss = val_loss_subtask / (self.num_low_level_tasks + 1)
                        running_val_loss += float(loss.item()) * images.size(0)

                model.train()
                train_loss = running_train_loss/len(train_loader)
                val_loss = running_val_loss / len(val_loader)
                metrics = self._get_evaluation_metric(true_labels_dict, predicted_labels_dict, scores_dict)
                # Appedning the losses
                train_losses.append(train_loss)
                val_losses.append(val_loss)

                print("Epoch:{}/{} - Training Loss:{:.6f} | Validation Loss: {:.6f}".format(
                    epoch+1, self.num_epochs, train_loss, val_loss))
                logging.info("Epoch:{}/{} - Training Loss:{:.6f} | Validation Loss: {:.6f}".format(
                    epoch+1, self.num_epochs, train_loss, val_loss))
                print("\tAccuracy | Precision | Sensitivity | Specificity | AUC")
                logging.info("\tAccuracy | Precision | Sensitivity | Specificity | AUC")
                for arg in metrics:
                    print("{}: {}".format(arg, metrics[arg]))
                    logging.info("{}: {}".format(arg, metrics[arg]))
                print("-" * 40)
                logging.info("-" * 40)

                # SAVE MODEL IF VALIDATION LOSS HAS DECREASED
                if val_loss <= valid_loss_min:
                    print("Validation loss decreased ({:.6f} --> {:.6f}).  Saving model ...".format(valid_loss_min, val_loss))
                    print("-"*40)
                    # Copy new weights
                    best_model_wts = True
                    # Save models
                    fold_wgt_path = "checkpoints/HSCNN_Fold-{}_Epoch-{}.pth".format(fold+1, epoch+1)
                    torch.save(model.state_dict(), fold_wgt_path)
                    best_epch_at = epoch+1
                    # Update minimum loss
                    valid_loss_min = val_loss
                else:
                    if best_model_wts:
                        # Update best weights
                        weights = torch.load("checkpoints/HSCNN_Fold-{}_Epoch-{}.pth".format(fold+1, best_epch_at))
                        model.load_state_dict(weights)
                        del weights

                # Delete to save memory
                del metrics, true_labels_dict, predicted_labels_dict, scores_dict, \
                    running_train_loss, running_val_loss

            # External test-set loop
            with torch.no_grad():
                runnning_test_loss = 0.0
                model.eval()
                predicted_labels_dict = {'label_1': [], 'label_2': [], 'label_3': [], 'label_4': [], 'label_5': [],
                                         'label_6': []}
                true_labels_dict = {'label_1': [], 'label_2': [], 'label_3': [], 'label_4': [], 'label_5': [],
                                    'label_6': []}
                scores_dict = {'label_1': [], 'label_2': [], 'label_3': [], 'label_4': [], 'label_5': [], 'label_6': []}

                for index, data in enumerate(test_loader):
                    images, labels = data['img'].to(self.device), data['label']
                    per_label_for_each_image = [labels[:, i].to(self.device, dtype=torch.long) for i in range(self.num_low_level_tasks + 1)]
                    output = model(images)
                    test_loss_subtask = 0
                    for i in range(self.num_low_level_tasks + 1):
                        criterion = criterions['criterion_{}'.format(i + 1)]
                        test_loss_subtask += criterion(output[i], per_label_for_each_image[i]) * loss_weights[i]
                        per_task_out = F.softmax(output[i].cpu(), dim=1)
                        top_ps, top_class = per_task_out.topk(1, dim=1)
                        predicted_labels_dict['label_{}'.format(i + 1)].extend(list(top_class.flatten().numpy()))
                        true_labels_dict['label_{}'.format(i + 1)].extend(list(per_label_for_each_image[i].cpu().numpy()))
                        scores_dict['label_{}'.format(i + 1)].extend(per_task_out.numpy().tolist())
                    loss = test_loss_subtask / (self.num_low_level_tasks + 1)
                    runnning_test_loss += float(loss.item()) * images.size(0)

            test_loss = runnning_test_loss / len(test_loader)
            metrics = self._get_evaluation_metric(true_labels_dict, predicted_labels_dict, scores_dict)
            aucs_across_folds.append(metrics["Malignancy"][-1])

            print("="*40)
            print("Test loss:", test_loss)
            logging.info("Test loss: {}".format(test_loss))
            print("-" * 40)
            print("\tAccuracy | Precision | Sensitivity | Specificity | AUC")
            logging.info("\tAccuracy | Precision | Sensitivity | Specificity | AUC")
            for arg in metrics:
                print("{}: {}".format(arg, metrics[arg]))
                logging.info("{}: {}".format(arg, metrics[arg]))
            print("=" * 40)

            plt.plot(train_losses, label='Training loss')
            plt.plot(val_losses, label='Validation loss')
            plt.xlabel('Loss')
            plt.ylabel('Epoch')
            plt.legend(frameon=False)
            plt.savefig('metrics_fold_{}_losses.png'.format(fold+1))
            plt.clf()

            fpr = dict()
            tpr = dict()
            roc_auc = dict()
            one_hot_true_lbls = np_utils.to_categorical(true_labels_dict['label_6'], num_classes=self.malignancy_class)
            y_scores = np.array(scores_dict['label_6'])
            # Computer FPR, TPR for two classes
            for i in range(self.malignancy_class):
                fpr[i], tpr[i], _ = roc_curve(one_hot_true_lbls[:, i], y_scores[:, i])
                roc_auc[i] = auc(fpr[i], tpr[i])
            """
            # Computer Micro FPR, TPR
            fpr["micro"], tpr["micro"], _ = roc_curve(one_hot_true_lbls.ravel(), y_scores.ravel())
            roc_auc["micro"] = auc(fpr["micro"], tpr["micro"])
            """
            # Computer Macro FPR, TPR
            all_fpr = np.unique(np.concatenate([fpr[i] for i in range(self.malignancy_class)]))
            # Then interpolate all ROC curves at this points
            mean_tpr = np.zeros_like(all_fpr)
            for i in range(self.malignancy_class):
                mean_tpr += interp(all_fpr, fpr[i], tpr[i])
            mean_tpr /= self.malignancy_class
            fpr["macro"] = all_fpr
            tpr["macro"] = mean_tpr
            roc_auc["macro"] = auc(fpr["macro"], tpr["macro"])

            plt.figure()
            plt.plot(fpr["macro"], tpr["macro"],
                     label='Average AUC Curve (area = {0:0.2f})'
                           ''.format(roc_auc["macro"]),
                     color='deeppink', linestyle=':', linewidth=4)
            colors = cycle(['aqua', 'darkorange'])
            for i, color in zip(range(self.malignancy_class), colors):
                plt.plot(fpr[i], tpr[i], color=color, lw=2,
                         label='ROC curve of class {0} (area = {1:0.2f})'
                               ''.format(i, roc_auc[i]))

            plt.plot([0, 1], [0, 1], 'k--', lw=2)
            plt.xlim([0.0, 1.0])
            plt.ylim([0.0, 1.05])
            plt.xlabel('False Positive Rate')
            plt.ylabel('True Positive Rate')
            plt.title('ROC - Malignancy')
            plt.legend(loc="lower right")
            plt.savefig('metrics_fold_{}_auc.png'.format(fold + 1))
            plt.clf()
            del train_losses, val_losses, accuracy_list, specificity_list, recall_list, f1_score_list, metrics, \
                true_labels_dict, predicted_labels_dict, scores_dict, runnning_test_loss, epoch_loss
        else:
            raise ValueError('Model must be instantiated and Trainloader/Validation loader cannot be empty!')

    def _get_index(self, input_index, num_fold):
        return input_index % num_fold

    def _load_pickle(self, input_file):
        with open(input_file, "rb") as f:
            pkl_file = pickle._Unpickler(f)
            pkl_file.encoding = 'latin1'
            pkl_file = pkl_file.load()
            return pkl_file

    def _merge_training_folds(self, path_to_training_fold, train_1_idx, train_2_idx):
        path_to_save = os.path.join(path_to_training_fold, "folds_merged")
        if not os.path.exists(path_to_save):
            os.makedirs(path_to_save)
        # Set up paths to training folds
        path_to_train_1 = path_to_training_fold+"fold_{}\\fold_{}_img.pkl".format(train_1_idx, train_1_idx)
        path_to_lbl_1 = path_to_training_fold + "fold_{}\\labels.csv".format(train_1_idx)
        path_to_train_2 = path_to_training_fold + "fold_{}\\fold_{}_img.pkl".format(train_2_idx, train_2_idx)
        path_to_lbl_2 = path_to_training_fold + "fold_{}\\labels.csv".format(train_2_idx)
        # Merge data and label
        img_1 = self._load_pickle(path_to_train_1)
        img_2 = self._load_pickle(path_to_train_2)
        new_img = np.concatenate((img_1, img_2), axis=0)
        lbl_1 = pd.read_csv(path_to_lbl_1)
        lbl_2 = pd.read_csv(path_to_lbl_2)
        frames = [lbl_1, lbl_2]
        df_concatenated = pd.concat(frames)

        # Save merged file and label
        pkl_name = "fold_{}_{}.pkl".format(train_1_idx, train_2_idx)
        full_path_lbl = os.path.join(path_to_save, "labels.csv")
        full_path_pkl = os.path.join(path_to_save, pkl_name)
        # Save Pickle file
        with open(full_path_pkl, 'wb') as file:
            pickle.dump(new_img, file)
        # Save csv file
        df_concatenated.to_csv(full_path_lbl, encoding='utf-8', index=False)
        return full_path_pkl, full_path_lbl

    def start_training(self, path_to_training_folds=None, num_folds=(3, 7)):
        if path_to_training_folds and num_folds:
            fold = 0
            for index in range(num_folds[0], num_folds[1]):
                # Set up path for folds
                test_idx, train_1_idx, train_2_idx, val_idx = self._get_index(index, 4), self._get_index(index+1, 4),\
                                                              self._get_index(index+2, 4), self._get_index(index+3, 4)

                print("Training fold: {}".format(fold+1))
                logging.info("Training fold: {}".format(fold+1))
                logging.info("Test Index: {}, Train Index:{}, Val Index:{}".format(test_idx, (train_1_idx, train_2_idx), val_idx))
                # Test set and labels
                path_to_test_fold = os.path.join(path_to_training_folds, 'fold_{}\\fold_{}_img.pkl'.format(test_idx, test_idx))
                path_to_test_lbl = os.path.join(path_to_training_folds, 'fold_{}\\labels.csv'.format(test_idx))
                # Val set and label
                path_to_val_fold = os.path.join(path_to_training_folds, 'fold_{}\\fold_{}_img.pkl'.format(val_idx, val_idx))
                path_to_val_lbl = os.path.join(path_to_training_folds, 'fold_{}\\labels.csv'.format(val_idx))
                # Get the merged training set with labels
                path_to_train_fold, path_to_train_lbl = self._merge_training_folds(path_to_training_folds,
                                                                 train_1_idx, train_2_idx)

                # Set up Dataset
                train_dataset = HscnnDataset(path_to_train_fold, path_to_train_lbl, num_low_level_tasks=3, transform=self._get_default_transforms())
                val_dataset = HscnnDataset(path_to_val_fold, path_to_val_lbl, num_low_level_tasks=3, transform=self._get_default_transforms())
                test_dataset = HscnnDataset(path_to_test_fold, path_to_test_lbl, num_low_level_tasks=3, transform=None)
                # Set up Dataloader
                train_loader = DataLoader(train_dataset, batch_size=self.batch_size, shuffle=True, drop_last=True)
                lbl_weights = self._class_weights(path_to_train_lbl)
                val_loader = DataLoader(val_dataset, batch_size=self.batch_size, shuffle=True, drop_last=True)
                test_loader = DataLoader(test_dataset, batch_size=self.batch_size, shuffle=True, drop_last=True)

                # Instantiate the model
                model = self._get_HSCNN(self.in_channel, self.num_low_level_tasks, self.low_level_outputs,
                                        self.malignancy_class, True)

                optimizer = optim.Adam(model.parameters(), lr=1e-3, weight_decay=1e-6)
                scheduler = optim.lr_scheduler.ReduceLROnPlateau(optimizer, patience=3, verbose=True)
                # Define criterions depending upon the number of tasks
                criterions = {}
                for i in range(self.num_low_level_tasks + 1):
                    weights = lbl_weights["task_{}_weights".format(i + 1)]
                    weights_tensor = torch.FloatTensor(weights).to(self.device)
                    #criterions['criterion_{}'.format(i + 1)] = nn.BCEWithLogitsLoss(pos_weight=weights_tensor)
                    criterions['criterion_{}'.format(i + 1)] = nn.CrossEntropyLoss(weight=weights_tensor)

                self._start_model(model, optimizer, criterions, scheduler, lbl_weights, self.loss_weights,
                                  train_loader, val_loader, test_loader, fold)
                fold += 1
                del train_loader, val_loader, test_loader, model, optimizer, scheduler, criterions
                # Delete the merged files
                os.remove(path_to_train_fold)
                os.remove(path_to_train_lbl)
                logging.info("Finished Fold {}".format(fold+1))
                return None
        else:
            return "Path to Training Fold is required!"


if __name__ == "__main__":
    # Hyper-param
    number_folds = (3, 7)
    in_channel = 1
    low_level_tasks = 5
    low_level_outputs = [2, 2, 2, 2, 2]
    loss_weights_each_task = [0.2, 0.1, 0.1, 0.2, 0.1, 1]
    malignancy_class = 2
    num_epochs = 200
    batch_size = 20
    val_per = 0.2

    # Call for Training
    path_to_folds = "training_fold"
    train_obj = TrainModel(in_channel, low_level_tasks, low_level_outputs, loss_weights_each_task, malignancy_class, num_epochs, batch_size, val_per)
    train_obj.start_training(path_to_folds, number_folds)




from setting import parse_opts 
from datasets.brains18 import BrainS18Dataset
import albumentations as A
from datasets.lungCT import LIDC_CT
from model import generate_model
import torch
import numpy as np
from torch import nn
from torch import optim
import pickle
from torch.optim import lr_scheduler
from torch.utils.data import DataLoader
import time
from utils.logger import log
from scipy import ndimage
import os
from tqdm import tqdm
from models.custom_loss import *
from build_unet_model import UNet


"""
Returns Intersection over Union score for ground truth and predicted masks
"""
def getIoU(y_true, y_pred):
    y_true_f = y_true.flatten()
    y_pred_f = y_pred.flatten()
    intersection = np.logical_and(y_true_f, y_pred_f).sum()
    union = np.logical_or(y_true_f, y_pred_f).sum()
    return (intersection + 1) * 1. / (union + 1)


"""
Training and Validation of Segmentation Model
"""
def train_and_validate(train_loader, val_loader, model, optimizer, scheduler, total_epochs, save_interval, save_folder, sets):
    # settings
    batches_per_epoch_train = len(train_loader)
    batches_per_epoch_val = len(val_loader)
    log.info('Train: {} epochs in total, {} batches per epoch'.format(total_epochs, batches_per_epoch_train))
    log.info('Val: {} epochs in total, {} batches per epoch'.format(total_epochs, batches_per_epoch_val))
    loss_dict = {'bce':[], 'dice':[], 'invt_dice':[]}

    # criteria to save best weight
    best_iou, total_ious = 0.0, []
    
    # define losses
    criterion1 = nn.BCEWithLogitsLoss()
    criterion2 = SoftDiceLoss()
    criterion3 = InvSoftDiceLoss()
    if not sets.no_cuda:
        criterion1 = nn.BCEWithLogitsLoss().cuda()
        criterion2 = SoftDiceLoss().cuda()
        criterion3 = InvSoftDiceLoss().cuda()
    
    # ------------------ #
    #     Training       #
    # ------------------ #
    model.train()
    train_time_sp = time.time()
    for epoch in range(total_epochs):
        log.info('Start epoch {}'.format(epoch))
        log.info('lr = {}'.format(scheduler.get_last_lr()))
        # running_total_loss = []

        bce_loss_over_epoch, dice_loss_over_epoch, ivt_loss_over_epoch = 0, 0, 0
        for batch_id, batch_data in enumerate(train_loader):
            batch_id_sp = epoch * batches_per_epoch_train
            volumes, label_masks = batch_data
            if not sets.no_cuda: 
                volumes = volumes.cuda()
                label_masks = label_masks.cuda()

            optimizer.zero_grad()
            out_masks = model(volumes)

            """
            # resize GT label
            [n, _, d, h, w] = out_masks.shape
            new_label_masks = np.zeros([n, _, d, h, w])
            for label_id in range(n):
                label_mask = label_masks[label_id]
                [ori_c, ori_d, ori_h, ori_w] = label_mask.shape
                label_mask = np.reshape(label_mask, [ori_d, ori_h, ori_w])
                scale = [d*1.0/ori_d, h*1.0/ori_h, w*1.0/ori_w]
                label_mask = ndimage.interpolation.zoom(label_mask, scale, order=0)
                new_label_masks[label_id, 0] = label_mask
            new_label_masks = torch.tensor(new_label_masks).type(torch.FloatTensor)
            if not sets.no_cuda:
                new_label_masks = new_label_masks.cuda()
            """
            
            # calculate loss
            # bce_loss, dice_loss, invt_dice_loss = criterion1(out_masks, new_label_masks), criterion2(out_masks, new_label_masks),\
            #                                     criterion3(out_masks, new_label_masks)

            bce_loss, dice_loss, invt_dice_loss = criterion1(out_masks, label_masks), criterion2(out_masks, label_masks),\
                                                criterion3(out_masks, label_masks)

            bce_loss_over_epoch += bce_loss.item()
            dice_loss_over_epoch += dice_loss.item()
            ivt_loss_over_epoch += invt_dice_loss.item()
            # running_total_loss.append(float(loss.item() * volumes.size(0)))
            loss = bce_loss + dice_loss + invt_dice_loss
            loss.backward()
            optimizer.step()

            # log metrics
            avg_batch_time = (time.time() - train_time_sp) / (1 + batch_id_sp)
            log.info(
                    'Batch: {}-{} ({}), loss_bce = {:.3f}, loss_dice = {:.3f}, loss_invt_dice = {:.3f}, total_loss = {:.3f}, avg_batch_time = {:.3f}'\
                    .format(epoch+1, batch_id, batch_id_sp, bce_loss.item(), dice_loss.item(), invt_dice_loss.item(), loss.item(), avg_batch_time))
            
        # append losses
        loss_dict['bce'].append(bce_loss_over_epoch/len(train_loader))
        loss_dict['dice'].append(dice_loss_over_epoch/len(train_loader))
        loss_dict['invt_dice'].append(ivt_loss_over_epoch/len(train_loader))

        if (epoch+1)%5 == 0:
            model_save_path = '{}_epoch_{}.pth.tar'.format(save_folder, epoch+1)
            model_save_dir = os.path.dirname(model_save_path)
            if not os.path.exists(model_save_dir):
                os.makedirs(model_save_dir)
            log.info('Save checkpoint: epoch = {}, batch_id = {}'.format(epoch+1, batch_id)) 
            torch.save({'epoch': epoch+1,
                        'batch_id': batch_id,
                        'state_dict': model.state_dict(),
                        'optimizer': optimizer.state_dict()},
                        model_save_path)
            loss_dict_path = '{}_loss_dict_epoch_{}.pkl'.format(save_folder, epoch+1)
            loss_save_dir = os.path.dirname(loss_dict_path)
            if not os.path.exists(loss_save_dir):
                os.makedirs(loss_save_dir)
            with open(loss_dict_path, 'wb') as f:
                pickle.dump(loss_dict, f)

        scheduler.step()
        # scheduler.step(np.mean(running_total_loss))
        # ------------------ #
        #     Validation     #
        # ------------------ #
        iou_over_epoch = 0
        with torch.no_grad():
            model.eval()
            for volumes, label_masks in tqdm(val_loader):
                if not sets.no_cuda:
                    volumes = volumes.cuda()
                
                out_masks = model(volumes)
                output_prob = torch.sigmoid(out_masks).detach().cpu().numpy()
                output_prob_thresh = (output_prob > 0.5) * 1

                """
                # resize GT label
                [n, _, d, h, w] = out_masks.shape
                new_label_masks = np.zeros([n, _, d, h, w])
                for label_id in range(n):
                    label_mask = label_masks[label_id]
                    [ori_c, ori_d, ori_h, ori_w] = label_mask.shape 
                    label_mask = np.reshape(label_mask, [ori_d, ori_h, ori_w])
                    scale = [d*1.0/ori_d, h*1.0/ori_h, w*1.0/ori_w]
                    label_mask = ndimage.interpolation.zoom(label_mask, scale, order=0)
                    new_label_masks[label_id, 0] = label_mask
                new_label_masks = torch.tensor(new_label_masks).to(torch.int64)
                """

                # iou = getIoU(new_label_masks, output_prob_thresh)
                iou = getIoU(label_masks, output_prob_thresh)
                iou_over_epoch += iou

            avg_iou = iou_over_epoch/len(val_loader)
            total_ious.append(avg_iou)
            log.info('Average IoU: {:.6f}'.format(avg_iou))

            # save model if average iou increases
            if avg_iou > best_iou:
                log.info("Average mask IoU increased: ({:.6f} --> {:.6f}).  Saving model ...".format(best_iou,\
                                                                                                  avg_iou))
                model_save_path = '{}_epoch_{}_best_weight_{:.6f}.pth.tar'.format(save_folder, epoch+1, avg_iou)
                model_save_dir = os.path.dirname(model_save_path)
                if not os.path.exists(model_save_dir):
                    os.makedirs(model_save_dir)
                torch.save({'epoch': epoch+1,
                            'batch_id': batch_id,
                            'state_dict': model.state_dict(),
                            'optimizer': optimizer.state_dict()},
                            model_save_path)
                best_iou = avg_iou


    print('Finished training')
    total_epochs_path = '{}_iou_over_{}_epochs.pkl'.format(save_folder, total_epochs)
    total_epochs_dir = os.path.dirname(total_epochs_path)
    if not os.path.exists(total_epochs_dir ):
        os.makedirs(total_epochs_dir)
    with open(total_epochs_path, 'wb') as f:
        pickle.dump(total_ious, f)
    exit()


if __name__ == '__main__':
    # settting
    sets = parse_opts()

    os.environ["CUDA_VISIBLE_DEVICES"]=str(sets.gpu_id[0])   
    model = UNet(in_nc=1, out_nc=1, nf=64).cuda()
    model = nn.DataParallel(model)

    # getting model
    torch.manual_seed(sets.manual_seed)
    # model, parameters = generate_model(sets) 

    # optimizer
    # if sets.ci_test:
    #     params = [{'params': parameters, 'lr': sets.learning_rate}]
    # else:
    #     params = [
    #             { 'params': parameters['base_parameters'], 'lr': sets.learning_rate }, 
    #             { 'params': parameters['new_parameters'], 'lr': sets.learning_rate*100 }
    #             ]

    # params = [{'params': parameters, 'lr': 1e-4}]
    # optimizer = torch.optim.SGD(params, momentum=0.9, weight_decay=1e-3)   
    # scheduler = optim.lr_scheduler.ExponentialLR(optimizer, gamma=0.99)

    optimizer = torch.optim.Adam(model.parameters(), lr=1e-4)
    scheduler = optim.lr_scheduler.ExponentialLR(optimizer, gamma=0.99)

    # getting data
    sets.phase = 'train'
    if sets.no_cuda:
        sets.pin_memory = False
    else:
        sets.pin_memory = True  

    # training transforms
    train_transform = A.Compose([
        A.HorizontalFlip(p=0.6),
        A.VerticalFlip(p=0.6),
        A.Transpose(p=0.6),
        A.RandomRotate90(p=0.6),
        A.OneOf([
            A.GaussNoise(p=0.6),
            A.RandomBrightnessContrast(p=0.5)
        ])
        # A.RandomCrop(width=64, height=64),
        # A.Resize(width=224, height=224),
    ])

    training_dataset = LIDC_CT('train', sets.data_root, sets.img_train_list, sets, transforms=train_transform)
    val_dataset = LIDC_CT('validation', sets.data_root, sets.img_val_list, sets, transforms=None)

    train_loader = DataLoader(training_dataset, batch_size=sets.batch_size, shuffle=True, num_workers=sets.num_workers, pin_memory=sets.pin_memory)
    val_loader = DataLoader(val_dataset, batch_size=8, shuffle=False, num_workers=sets.num_workers, pin_memory=sets.pin_memory)
    
    # data, label = iter(train_loader).next()
    # print('data shape:', data.shape)
    # print('label shape:', label.shape)

    # start training
    train_and_validate(train_loader, val_loader, model, optimizer, scheduler, total_epochs=sets.n_epochs, save_interval=sets.save_intervals, save_folder=sets.save_folder, sets=sets) 



import numpy as np
import os
from sklearn.model_selection import KFold


def main(path_to_uids):
    with open(path_to_uids) as f:
        path_to_save_folds = '/'.join(path_to_uids.split('/')[:-1])
        print('Saving folds uids at:', path_to_save_folds)
        lines = np.array(f.readlines())
        cv = KFold(n_splits=5, random_state=1, shuffle=True)
        for loop_idx, (train_index, valid_index) in enumerate(cv.split(lines)):
            fold_name_train = os.path.join(path_to_save_folds, "ucla_train_uids_fold_{}.txt".format(loop_idx+1))
            fold_name_validation = os.path.join(path_to_save_folds, "ucla_valid_uids_fold_{}.txt".format(loop_idx+1))
            train_uids = lines.take(train_index).tolist()
            validation_uids = lines.take(valid_index).tolist()
            # write out uids
            with open(fold_name_train, 'w') as train_file:
                train_file.write(''.join(train_uids))
            with open(fold_name_validation, 'w') as valid_file:
                valid_file.write(''.join(validation_uids))
    print('All folds saved!')


if __name__ == "__main__":
    path_to_uids = "/workspace/NormGAN/uids_to_split/ucla_all_uids.txt" 
    main(path_to_uids)


import os
import nrrd
import numpy as np
import matplotlib.pyplot as plt
import json
import cv2
import csv
from PIL import Image
import random


def rescale_img_to_rgb(img, need_clip=True):
    if need_clip:
        img = np.clip(img, 0, 1500)
    img = (img - img.min()) / (img.max() - img.min())
    img *= 255.0
    img = img.astype(np.uint8)
    return img

def generate_nodule_slices(ALLOWED_CND, test_set_annot):
    with open(test_set_annot, 'r') as file:
        nastaran_nodule_annot = json.load(file)
    print('Number of Nastaran nodule cases:', len(nastaran_nodule_annot.keys()))
    num_nastaran_nodule = 0
    for uid in nastaran_nodule_annot:
        center = nastaran_nodule_annot[uid]
        num_nastaran_nodule += len(center)
    print('Number of nodules in nastaran:', num_nastaran_nodule)

    # extract for each condition
    for cnd in ALLOWED_CND:
        if save_csv_annot:
            annot_csv = []
        print('Processing nodule condition:', cnd)
        save_path = os.path.join(path_to_save, cnd)
        if not os.path.exists(save_path):
            os.makedirs(save_path)
    
        file_path_to_read = os.path.join(path_to_cases, cnd)
        files = os.listdir(file_path_to_read)
        for case in nastaran_nodule_annot:
            file_name = case + '.nrrd'
            if file_name in files:
                img_path = os.path.join(file_path_to_read, file_name)
                data, header = nrrd.read(img_path)
                data = data.transpose((2, 1, 0))
                coords = nastaran_nodule_annot[case]
                coords = np.array([[int(x),int(y),int(z)] for x,y,z in coords])
                for annot in coords:
                    x, y, z = annot
                    xs, xe = x - 25, x + 25
                    ys, ye = y - 25, y + 25
                    try:
                        nodule_slice = data[z]
                        rgb_slice = Image.fromarray(rescale_img_to_rgb(nodule_slice)).convert("RGB")
                        nodule_name = 'nod_{}_{}.png'.format(case, z)
                        nodule_save_path = os.path.join(save_path, nodule_name)
                        rgb_slice.save(nodule_save_path)
                        if save_csv_annot:
                            annot_csv.append([nodule_name, xs, ys, xe, ye, 'nodule'])
                    except Exception as e:
                        print('{}, Skipping....'.format(e))
        
        if save_csv_annot:
            save_path = os.path.join(path_to_save, '{}.csv'.format(cnd))
            with open(save_path, 'w') as f:
                write = csv.writer(f)
                write.writerows(annot_csv)

    print('All Nodule Slices Extracted!')


def generate_nonod_slices(ALLOWED_CND, nonod_annot):
    with open(nonod_annot, 'r') as file:
        nas_nonodule_annot = json.load(file)
    # extract for each condition
    for cnd in ALLOWED_CND:
        print('Processing non-nodule condition:', cnd)
        save_path = os.path.join(path_to_save, cnd)
        file_path_to_read = os.path.join(path_to_cases, cnd)
        if not os.path.exists(save_path):
            os.makedirs(save_path)

        files = os.listdir(file_path_to_read)
        for case in nas_nonodule_annot:
            file_name = case + '.nrrd'
            if file_name in files:
                img_path = os.path.join(file_path_to_read, file_name)
                data, header = nrrd.read(img_path)
                data = data.transpose((2, 1, 0))
                coords = nas_nonodule_annot[case]
                for annot in coords:
                    z = (int(annot[0]) + int(annot[1]))//2
                    try:
                        data_annot = data[z, :, :]
                        clipped = np.clip(data_annot, 0, 1500)
                        clipped = np.true_divide(clipped, 1500)
                        clipped = (clipped*255.0).astype(np.uint8)
                        clipped_img = Image.fromarray(clipped)
                        nodule_name = 'nonod_{}_{}.png'.format(case, z)
                        nodule_save_path = os.path.join(save_path, nodule_name)
                        clipped_img.save(nodule_save_path)
                        nonodannot_csv.append([nodule_save_path])
                    except Exception as e:
                        print('{}, Skipping....'.format(e))             
    print('All Non-nodule Slices Extracted!')


if __name__ == '__main__':
    ALLOWED_CND = ['k1_d10_st1.0', 'k1_d25_st1.0', 'k1_d100_st1.0', 'k2_d10_st1.0', 'k2_d100_st1.0', 'k3_d10_st1.0', 'k3_d25_st1.0', 'k3_d100_st1.0']
    path_to_cases = '/dingo_data_leihao/nastaran_test_h5'
    save_csv_annot = True
    
    # ALLOWED_CND = ['test_cases-Nas-k1_d10_st1.0', 'test_cases-Nas-k1_d25_st1.0', 'test_cases-Nas-k1_d100_st1.0',\
    #                 'test_cases-Nas-k2_d10_st1.0', 'test_cases-Nas-k3_d10_st1.0', 'test_cases-Nas-k3_d25_st1.0',\
    #                 'test_cases-Nas-k3_d100_st1.0']
    # path_to_cases = '/workspace/cNormGAN-AC/results/nastaran_test_cases/HM-EQMergedData_Ref-k2d100'

    test_set_annot = '/datasets/nodule_annotations/ucla_nastaran_test_set_st1.0.json'
    path_to_save = '/datasets/Nas_TaskData/task_detection_512_512/Unnormalized'
    subset_folder = '' # test_st1.0_norm
    
    # extract nodule slices
    generate_nodule_slices(ALLOWED_CND, test_set_annot)
    # generate_nonod_slices(ALLOWED_CND, nonod_annot)







import copy
from .rcnn_nms import rcnn_encode
import time
import random
import numpy as np
import torch
from torch.autograd import Variable
try:
    from lung_analysis_pipeline.NoduleDetect_SANet.utils.pybox import *
except ImportError:
    print('Warning: C++ module import failed! This should only happen in deployment')
    from lung_analysis_pipeline.NoduleDetect_SANet.utils.util import py_nms as torch_nms
    from lung_analysis_pipeline.NoduleDetect_SANet.utils.util import py_box_overlap as torch_overlap

score = 1

def add_truth_box_to_proposal(cfg, proposal, b, truth_box, truth_label, score=1):
    if len(truth_box) !=0:
        truth = np.zeros((len(truth_box), 8),np.float32)
        truth[:, 0] = b
        truth[:, 2:8] = truth_box
        truth[:, 1] = score #1  #
    else:
        truth = np.zeros((0, 8),np.float32)

    sampled_proposal = np.vstack([proposal,truth])
    return sampled_proposal


def make_one_rcnn_target(cfg, input, proposal, truth_box, truth_label):
    sampled_proposal = torch.zeros((0, 8)).float().cuda()
    sampled_label = torch.zeros((0, 1)).long().cuda()
    sampled_assign = np.zeros((0, 1), dtype=np.int32) - 1
    sampled_target = torch.zeros((0, 6)).float().cuda()

    # Even if there is no ground truth box in this batch
    if len(proposal) == 0:
        return sampled_proposal, sampled_label, sampled_assign, sampled_target

    if len(truth_box) == 0:
        num_bg = min(len(proposal), cfg['rcnn_train_batch_size'])
        bg_length = len(proposal)
        bg_index = np.arange(len(proposal))
        bg_index = bg_index[
            np.random.choice(bg_length, size=num_bg, replace=bg_length<num_bg)
        ]
        sampled_proposal = proposal[bg_index]
        sampled_proposal = torch.from_numpy(sampled_proposal).cuda()
        sampled_label = torch.zeros((num_bg)).long().cuda()

        return sampled_proposal, sampled_label, sampled_assign, sampled_target 

    _, depth, height, width = input.size()
    num_proposal = len(proposal)
    box = proposal[:, 2:8]

    # Determine positive or negative purely based on threshold
    # Since the GT box has been added to proposal, it is gauranteed that
    # each ground truth would have one proposal
    overlap = torch_overlap(box, truth_box)
    argmax_overlap = np.argmax(overlap,1)
    max_overlap = overlap[np.arange(num_proposal),argmax_overlap]

    fg_index = np.where(max_overlap >= cfg['rcnn_train_fg_thresh_low'])[0]
    bg_index = np.where(max_overlap <  cfg['rcnn_train_bg_thresh_high'])[0]

    # sampling for class balance
    num_class = cfg['num_class']
    num = cfg['rcnn_train_batch_size']
    num_fg = int(np.round(cfg['rcnn_train_fg_fraction'] * cfg['rcnn_train_batch_size']))

    fg_length = len(fg_index)
    bg_length = len(bg_index)
    #print(fg_inds_length)

    sampled_assign = argmax_overlap[fg_index]

    # Need to consider four cases, corner cases
    if fg_length > 0 and bg_length > 0:
        idx = []
        idx = random.sample(range(len(fg_index)), min(num_fg, len(fg_index)))
        
        fg_index = fg_index[idx]
        num_fg = len(fg_index)

        num_bg  = num - num_fg
        bg_index = bg_index[
            np.random.choice(bg_length, size=num_bg, replace=bg_length<num_bg)
        ]
    elif fg_length > 0:  #no bgs
        idx = []
        idx = random.sample(range(len(fg_index)), min(num_fg, len(fg_index)))
        
        fg_index = fg_index[idx]
        num_fg = len(fg_index)
        num = num_fg
        num_bg = 0

    elif bg_length > 0:  #no fgs
        print('[RCNN] No fgs')
        print(truth_box)
        print('---------------------------')
        print(proposal)

        num_fg = 0
        num_bg = num
        bg_index = bg_index[
            np.random.choice(bg_length, size=num_bg, replace=bg_length<num_bg)
        ]
        num_fg_proposal = 0
    else:
        # no bgs and no fgs?
        # raise NotImplementedError
        print('[RCNN] No bgs or fgs')
        print(truth_box)
        print('---------------------------')
        print(proposal)
        num_fg   = 0
        num_bg   = num
        bg_index = np.random.choice(num_proposal, size=num_bg, replace=num_proposal<num_bg)

    assert ((num_fg+num_bg)== num)

    # selecting both fg and bg
    index = np.concatenate([fg_index, bg_index], 0)
    sampled_proposal = np.take(proposal, index, axis=0)

    # label
    sampled_assign = np.take(argmax_overlap, index)
    sampled_label = np.take(truth_label, sampled_assign)
    
    sampled_label[num_fg:] = 0   # Clamp labels for the background to 0
    sampled_assign[num_fg:] = -1 # Clamp label assignments for the background to -1

    # bounding box regression terms
    if num_fg>0:
        target_truth_box = truth_box[sampled_assign[:num_fg], :]
        if len(target_truth_box.shape) < 2: # one dimension lost after slicing
            target_truth_box = target_truth_box[np.newaxis, ...]
        target_box = sampled_proposal[:num_fg,:][:, 2:8]
        sampled_target = rcnn_encode(target_box, target_truth_box, cfg['box_reg_weight'])

    sampled_target   = Variable(torch.from_numpy(sampled_target)).float().cuda()
    sampled_label    = Variable(torch.from_numpy(sampled_label)).long().cuda()
    sampled_proposal = Variable(torch.from_numpy(sampled_proposal)).cuda()

    return sampled_proposal, sampled_label, sampled_assign, sampled_target


def make_rcnn_target(cfg, mode, inputs, proposals, truth_boxes, truth_labels):
    truth_boxes = copy.deepcopy(truth_boxes)
    truth_labels = copy.deepcopy(truth_labels)
    batch_size = len(inputs)
    for b in range(batch_size):
        index = np.where(truth_labels[b] > 0)[0]
        truth_boxes [b] = truth_boxes [b][index]
        truth_labels[b] = truth_labels[b][index]

    proposals = proposals.cpu().data.numpy()
    sampled_proposals = []
    sampled_labels = []
    sampled_assigns = []
    sampled_targets = []
    sampled_masks = []

    batch_size = len(truth_boxes)
    for b in range(batch_size):
        input = inputs[b]
        truth_box = truth_boxes[b]
        truth_label = truth_labels[b]

        if len(proposals) == 0:
            proposal = np.zeros((0, 8),np.float32)
        else:
            proposal = proposals[proposals[:,0] == b]

        # Add ground truth box to proposal, so that even if the RPN branch fails to find something,
        # we can still get classification branch to work
        proposal = add_truth_box_to_proposal(cfg, proposal, b, truth_box, truth_label)

        sampled_proposal, sampled_label, sampled_assign, sampled_target = \
           make_one_rcnn_target(cfg, input, proposal, truth_box, truth_label)

        sampled_proposals.append(sampled_proposal)
        sampled_labels.append(sampled_label)
        sampled_assigns.append(sampled_assign)
        sampled_targets.append(sampled_target)

    sampled_proposals = torch.cat(sampled_proposals, 0)
    sampled_labels = torch.cat(sampled_labels, 0)
    sampled_targets = torch.cat(sampled_targets, 0)
    sampled_assigns = np.hstack(sampled_assigns)

    return sampled_proposals, sampled_labels, sampled_assigns, sampled_targets

 



import os
import os.path as osp
import logging
from collections import OrderedDict
import json


"""
Reads the json configuration file and loads the dictionary into an 'OrderedDict()'. 
e.g. OrderedDict([('name', '3DGAN_A01'), ('use_tb_logger', True), ...])
"""
def parse(opt_path, is_train=True):
    # remove comments starting with '//'
    json_str = ''
    with open(opt_path, 'r') as f:
        for line in f:
            line = line.split('//')[0] + '\n'
            json_str += line
    opt = json.loads(json_str, object_pairs_hook=OrderedDict)
    opt['is_train'] = is_train
    scale = opt['scale']
    # datasets
    for phase, dataset in opt['datasets'].items():
        phase = phase.split('_')[0]
        dataset['phase'] = phase
        dataset['scale'] = scale
    # path
    for key, path in opt['path'].items():
        if path and key in opt['path']:
            opt['path'][key] = os.path.expanduser(path)
    if is_train:
        experiments_root = os.path.join(opt['path']['root'], 'experiments_paper', opt['name'])
        opt['path']['experiments_root'] = experiments_root
        opt['path']['models'] = os.path.join(experiments_root, 'models')
        opt['path']['training_state'] = os.path.join(experiments_root, 'training_state')
        opt['path']['log'] = experiments_root
        opt['path']['val_images'] = os.path.join(experiments_root, 'val_images')

        # change some options for debug mode
        if 'debug' in opt['name']:
            opt['train']['val_freq'] = 10
            opt['logger']['print_freq'] = 1
            opt['logger']['save_checkpoint_freq'] = 50
            opt['train']['lr_decay_iter'] = 10
    else:  # test
        results_root = opt['path']['results_dir']
        opt['path']['results_root'] = results_root
        opt['path']['log'] = results_root

    # network
    opt['network_G']['scale'] = scale
    # export CUDA_VISIBLE_DEVICES
    gpu_list = ','.join(str(x) for x in opt['gpu_ids']) if opt['gpu_ids'] else ""
    os.environ['CUDA_VISIBLE_DEVICES'] = gpu_list
    print('export CUDA_VISIBLE_DEVICES=' + gpu_list)
    return opt


class NoneDict(dict):
    def __missing__(self, key):
        return None


"""
Converts the 'OrderedDict()'to NoneDict, which return None for missing key
"""
def dict_to_nonedict(opt):
    if isinstance(opt, dict):
        new_opt = dict()
        for key, sub_opt in opt.items():
            new_opt[key] = dict_to_nonedict(sub_opt)
        return NoneDict(**new_opt)
    elif isinstance(opt, list):
        return [dict_to_nonedict(sub_opt) for sub_opt in opt]
    else:
        return opt

"""
Converts dictionary into readable string
"""
def dict2str(opt, indent_l=1):
    msg = ''
    for k, v in opt.items():
        if isinstance(v, dict):
            msg += ' ' * (indent_l * 2) + k + ':[\n'
            msg += dict2str(v, indent_l + 1)
            msg += ' ' * (indent_l * 2) + ']\n'
        else:
            msg += ' ' * (indent_l * 2) + k + ': ' + str(v) + '\n'
    return msg


"""
Check resume states and pretrain_model paths
"""
def check_resume(opt):
    logger = logging.getLogger('base')
    if opt['path']['resume_state']:
        if opt['path']['pretrain_model_G'] or opt['path']['pretrain_model_D']:
            logger.warning('pretrain_model path will be ignored when resuming training.')

        state_idx = osp.basename(opt['path']['resume_state']).split('.')[0]
        opt['path']['pretrain_model_G'] = osp.join(opt['path']['models'],
                                                   '{}_G.pth'.format(state_idx))
        logger.info('Set [pretrain_model_G] to ' + opt['path']['pretrain_model_G'])
        if 'gan' in opt['model']:
            opt['path']['pretrain_model_D'] = osp.join(opt['path']['models'],
                                                       '{}_D.pth'.format(state_idx))
            logger.info('Set [pretrain_model_D] to ' + opt['path']['pretrain_model_D'])



import numpy as np
import torch
from torch.utils.data import Dataset
import os, imageio
from scipy.ndimage import zoom
import warnings
from scipy.ndimage.interpolation import rotate
import math
import time
import nrrd
import pandas as pd
import SimpleITK as sitk
from utils.util import normalize, truncate_HU_uint8


class BboxReader(Dataset):
    def __init__(self, data_dir, set_name, cfg, mode='train'):
        self.mode = mode
        print('mode:', self.mode)
        self.cfg = cfg
        self.r_rand = cfg['r_rand_crop']
        self.augtype = cfg['augtype']
        self.pad_value = cfg['pad_value']
        self.data_dir = data_dir
        self.stride = cfg['stride']
        self.blacklist = cfg['blacklist']
        self.set_name = set_name
        sizelim = 0
        sizelim2 = 10
        sizelim3 = 20

        labels = []
        with open(self.set_name, "r") as f:
            self.filenames = f.read().splitlines()

        if mode != 'test':
            self.filenames = [f for f in self.filenames if (f not in self.blacklist)]
        if self.mode in ['train', 'val']:
            csv_dir = cfg['train_anno']
        else:
            csv_dir = cfg['test_anno']

        annos_all = pd.read_csv(csv_dir)
        # collecting annotations
        for fn in self.filenames:
            annos = annos_all[annos_all['pid'] == fn]            
            temp_annos = []
            if len(annos) > 0:
                for index in range(len(annos)):
                    anno = annos.iloc[index]
                    temp_annos.append([anno['zmin'], anno['zmax'], anno['ymin'], anno['ymax'], anno['xmin'], anno['xmax']])
            l = np.array(temp_annos)
            if np.all(l==0):
                l=np.array([])
            labels.append(l)

        self.sample_bboxes = labels
        if self.mode in ['train', 'val']:
            self.bboxes = []
            for i, l in enumerate(labels):
                if len(l) > 0 :
                    for t in l:
                        # diameter = max(t[5] - t[4], t[3] - t[2])
                        # if diameter>sizelim:
                        #     self.bboxes.append([np.concatenate([[i],t])])
                        # if diameter>sizelim2:
                        #     self.bboxes+=[[np.concatenate([[i],t])]]*2
                        # if diameter>sizelim3:
                        #     self.bboxes+=[[np.concatenate([[i],t])]]*4
                        self.bboxes.append([np.concatenate([[i],t])])
            self.bboxes = np.concatenate(self.bboxes, axis=0).astype(np.float32)
            # print('self.bboxes:', self.bboxes)
        self.crop = Crop(cfg)

    def load_itk_image(self, filename):
        itkimage = sitk.ReadImage(filename)
        numpyImage = sitk.GetArrayFromImage(itkimage)
        numpyImage[numpyImage == -2000] = 0
        numpyOrigin = np.array(list(reversed(itkimage.GetOrigin())))
        numpySpacing = np.array(list(reversed(itkimage.GetSpacing())))
        return numpyImage, numpyOrigin, numpySpacing

    def __getitem__(self, idx):
        t = time.time()
        np.random.seed(int(str(t % 1)[2:7]))  # seed according to time
        is_random_img  = False
        if self.mode in ['train', 'val']:
            if idx >= len(self.bboxes):
                is_random_crop = True
                idx = idx % len(self.bboxes)
                is_random_img = np.random.randint(2)
            else:
                is_random_crop = False
        else:
            is_random_crop = False
            # print('is_random_crop', False)

        if self.mode in ['train', 'val']:
            print('Entering in train or val')
            if not is_random_img:
                bbox = self.bboxes[idx]
                filename = self.filenames[int(bbox[0])]
                imgs = np.load(os.path.join(self.data_dir, '%s_zoom.npy' % (filename)))

                # lines = os.listdir(os.path.join(self.data_dir.replace('full', 'img'), filename))
                # lines = sorted(lines)
                # slice_files = [os.path.join(self.data_dir.replace('full', 'img'), filename, s) for s in lines]
                # slices = [imageio.imread(s) for s in slice_files]
                # imgs = np.array(slices)
                # imgs = imgs[np.newaxis, ...]

                bboxes = self.sample_bboxes[int(bbox[0])]
                isScale = self.augtype['scale'] and (self.mode=='train')
                sample, target, bboxes, coord = self.crop(imgs, bbox[1:], bboxes,isScale,is_random_crop)
                if self.mode == 'train' and not is_random_crop:
                     sample, target, bboxes = augment(sample, target, bboxes, do_flip = self.augtype['flip'], 
                                                             do_rotate=self.augtype['rotate'], do_swap = self.augtype['swap'])
            else:
                randimid = np.random.randint(len(self.filenames))
                filename = self.filenames[randimid]
                imgs = np.load(os.path.join(self.data_dir, '%s_zoom.npy' % (filename)))

                # lines = os.listdir(os.path.join(self.data_dir.replace('full', 'img'), filename))
                # lines = sorted(lines)
                # slice_files = [os.path.join(self.data_dir.replace('full', 'img'), filename, s) for s in lines]
                # slices = [imageio.imread(s) for s in slice_files]
                # imgs = np.array(slices)
                # imgs = imgs[np.newaxis, ...]

                bboxes = self.sample_bboxes[randimid]
                isScale = self.augtype['scale'] and (self.mode=='train')
                sample, target, bboxes, coord = self.crop(imgs, [], bboxes,isScale=False,isRand=True)

            if sample.shape[1] != self.cfg['crop_size'][0] or sample.shape[2] != \
                self.cfg['crop_size'][1] or sample.shape[3] != self.cfg['crop_size'][2]:
                print(filename, sample.shape)

            sample = (sample.astype(np.float32)-128)/128
            bboxes = fillter_box(bboxes, self.cfg['crop_size'])
            bboxes = corner_form_to_center_form(bboxes, self.cfg['bbox_border'])
            bboxes = np.array(bboxes)
            truth_labels = bboxes[:, -1]
            truth_bboxes = bboxes[:, :-1]

            return [torch.from_numpy(sample).float(), truth_bboxes, truth_labels]

        if self.mode in ['eval']:
            filename = self.filenames[idx]
            # image = np.load(os.path.join(self.data_dir, '%s_zoom.npy' % (filename)))
            numpyImage, numpyOrigin, numpySpacing = self.load_itk_image(os.path.join(self.data_dir, filename))
            numpyImage = bound_image(numpyImage, lower_bound=-1200, upper_bound=600)
            # convert to rgb range
            numpyImage = truncate_HU_uint8(numpyImage)

            # numpyImage = (numpyImage - numpyImage.min()) / (numpyImage.max() - numpyImage.min())
            # numpyImage = (numpyImage * 255).astype('uint8')
            # lines = os.listdir(os.path.join(self.data_dir.replace('full', 'img'), filename))
            # lines = sorted(lines)
            # slice_files = [os.path.join(self.data_dir.replace('full', 'img'), filename, s) for s in lines]
            # slices = [imageio.imread(s) for s in slice_files]
            # imgs = np.array(slices)
            # image = imgs[np.newaxis, ...]
    
            # original_image = numpyImage
            image = pad2factor(numpyImage)
            image = np.expand_dims(image, 0)
            bboxes = self.sample_bboxes[idx]

            bboxes = corner_form_to_center_form(bboxes, self.cfg['bbox_border'])
            bboxes = np.array(bboxes)
            truth_labels = bboxes[:, -1]
            truth_bboxes = bboxes[:, :-1]

            # input = normalize(image.astype(np.float32))
            input = (image.astype(np.float32) - 128.) / 128.


            return [torch.from_numpy(input).float(), truth_bboxes, truth_labels]


    def __len__(self):
        if self.mode == 'train':
            return int(len(self.bboxes) / (1-self.r_rand))//self.cfg['batch_size']*self.cfg['batch_size']
        elif self.mode =='val':
            return len(self.bboxes)//self.cfg['batch_size']*self.cfg['batch_size']
        else:
            return len(self.filenames)


def bound_image(img, lower_bound=-1000, upper_bound=500, shift_pixels=0):
    image_data = img.astype('float32')
    image_data[image_data < lower_bound] = lower_bound
    image_data[image_data > upper_bound] = upper_bound
    shifted_pixels = image_data + shift_pixels
    shifted_pixels = shifted_pixels.astype(np.int16)
    return shifted_pixels


def corner_form_to_center_form(boxes, border):
    # print('converting bbox for corner to center')
    # print('bboxes are:', boxes)
    # print('border:', border)
    bboxes = []
    for box in boxes:
        """
        print('[({}+{})/2., ({}+{})/2., ({}+{})/2., ({}-{}+1+{}), ({}-{}+1+{}), ({}-{}+1+{})]'.format(box[0], box[1], box[2], box[3],
                                                                                        box[4], box[5], box[1], box[0], border,
                                                                                        box[3], box[2], border,
                                                                                        box[5], box[4], border))
        """
        bboxes.append([(box[0] + box[1]) / 2.,
                       (box[2] + box[3]) / 2.,
                       (box[4] + box[5]) / 2.,
                       box[1] - box[0] + 1 + border,
                       box[3] - box[2] + 1 + border,
                       box[5] - box[4] + 1 + border,
                       1])

    # print('converted boxes:', bboxes)  
    return bboxes

def pad2factor(image, factor=32, pad_value=0):
    # print('Inside pad2factor')
    # print('-'*40)
    depth, height, width = image.shape
    # print('depth, height, width:', depth, height, width)
    d = int(math.ceil(depth / float(factor))) * factor
    # print("d = int(math.ceil({}/{})) * {} = {}".format(depth, factor, factor, d))
    h = int(math.ceil(height / float(factor))) * factor
    # print("h = int(math.ceil({}/{})) * {} = {}".format(height, factor, factor, h))
    w = int(math.ceil(width / float(factor))) * factor
    # print("w = int(math.ceil({}/{})) * {} = {}".format(width, factor, factor, w))

    pad = []
    pad.append([0, d - depth])
    pad.append([0, h - height])
    pad.append([0, w - width])
    # print('adding pad: [{}-{}, {}-{}, {}-{}] = [{}, {}, {}]'.format(d, depth, h, height, w, width, d-depth, h-height, w-width))

    # pad = []
    # pad.append([0, w - width + height-depth])
    # pad.append([0, h - height])
    # pad.append([0, w - width])

    image = np.pad(image, pad, 'constant', constant_values=pad_value)

    return image



def fillter_box(bboxes, size):
    res = []
    for box in bboxes:
        if np.all(box[:6] > 0) and np.all(box[:6]  < size[0]):
            res.append(box)
    return np.array(res)

def augment(sample, target, bboxes, do_flip = True, do_rotate=True, do_swap = True):
    #  angle1 = np.random.rand()*180
    if do_rotate:
        validrot = False
        counter = 0
        while not validrot:
            newtarget = np.copy(target)
            angle1 = np.random.rand()*180
            size = np.array(sample.shape[2:4]).astype('float')
            rotmat = np.array([[np.cos(angle1/180*np.pi),-np.sin(angle1/180*np.pi)],[np.sin(angle1/180*np.pi),np.cos(angle1/180*np.pi)]])
            newtarget[1:3] = np.dot(rotmat,target[1:3]-size/2)+size/2
            if np.all(newtarget[:3]>target[3]) and np.all(newtarget[:3]< np.array(sample.shape[1:4])-newtarget[3]):
                validrot = True
                target = newtarget
                sample = rotate(sample,angle1,axes=(2,3),reshape=False)
                for box in bboxes:
                    box[1:3] = np.dot(rotmat,box[1:3]-size/2)+size/2
            else:
                counter += 1
                if counter ==3:
                    break
    if do_swap:
        if sample.shape[1]==sample.shape[2] and sample.shape[1]==sample.shape[3]:
            axisorder = np.random.permutation(3)
            sample = np.transpose(sample,np.concatenate([[0],axisorder+1]))
            target[:3] = target[:3][axisorder]
            bboxes[:,:3] = bboxes[:,:3][:,axisorder]

    if do_flip:
        # flipid = np.array([np.random.randint(2),np.random.randint(2),np.random.randint(2)])*2-1
        flipid = np.array([1,np.random.randint(2),np.random.randint(2)])*2-1
        sample = np.ascontiguousarray(sample[:,::flipid[0],::flipid[1],::flipid[2]])
        # for ax in range(3):
        #     if flipid[ax]==-1:
        #         target[ax] = np.array(sample.shape[ax+1])-target[ax]
        #         bboxes[:,ax]= np.array(sample.shape[ax+1])-bboxes[:,ax]
        for i in range(3):
            if flipid[i]==-1:
                tem = [0, 2, 4]
                ax = tem[i]
                target[ax] = np.array(sample.shape[i+1])-target[ax]
                bboxes[:,ax]= np.array(sample.shape[i+1])-bboxes[:,ax]
                target[ax+1] = np.array(sample.shape[i + 1]) - target[ax+1]
                bboxes[:, ax+1] = np.array(sample.shape[i + 1]) - bboxes[:, ax+1]
    return sample, target, bboxes

class Crop(object):
    # print('Inside Crop Class')
    # print('-'*40)
    def __init__(self, config):
        self.crop_size = config['crop_size']
        self.bound_size = config['bound_size']
        self.stride = config['stride']
        self.pad_value = config['pad_value']
    def __call__(self, imgs, target, bboxes,isScale=False,isRand=False):
        if isScale:
            radiusLim = [8.,120.]
            scaleLim = [0.75,1.25]
            scaleRange = [np.min([np.max([(radiusLim[0]/target[3]),scaleLim[0]]),1])
                         ,np.max([np.min([(radiusLim[1]/target[3]),scaleLim[1]]),1])]
            scale = np.random.rand()*(scaleRange[1]-scaleRange[0])+scaleRange[0]
            crop_size = (np.array(self.crop_size).astype('float')/scale).astype('int')
        else:
            crop_size=self.crop_size
        bound_size = self.bound_size
        target = np.copy(target)
        bboxes = np.copy(bboxes)

        start = []
        for i in range(3):
            # start.append(int(target[i] - crop_size[i] / 2))
            if not isRand:
                tem = [0,2,4]
                j = tem[i]
                # r = target[3] / 2
                # s = np.floor(target[i] - r)+ 1 - bound_size
                # e = np.ceil (target[i] + r)+ 1 + bound_size - crop_size[i]
                s = target[j] - bound_size
                e = target[j+1] + bound_size - crop_size[i]
            else:
                s = np.max([imgs.shape[i+1]-crop_size[i]/2,imgs.shape[i+1]/2+bound_size])
                e = np.min([crop_size[i]/2, imgs.shape[i+1]/2-bound_size])
                target = np.array([np.nan,np.nan,np.nan,np.nan])
            if s>e:
                start.append(np.random.randint(e,s))#!
            else:
                start.append(int(target[i])-int(crop_size[i]/2)+np.random.randint(-bound_size/2,bound_size/2))


        normstart = np.array(start).astype('float32')/np.array(imgs.shape[1:])-0.5
        normsize = np.array(crop_size).astype('float32')/np.array(imgs.shape[1:])
        xx,yy,zz = np.meshgrid(np.linspace(normstart[0],normstart[0]+normsize[0],int(self.crop_size[0]/self.stride)),
                           np.linspace(normstart[1],normstart[1]+normsize[1],int(self.crop_size[1]/self.stride)),
                           np.linspace(normstart[2],normstart[2]+normsize[2],int(self.crop_size[2]/self.stride)),indexing ='ij')
        coord = np.concatenate([xx[np.newaxis,...], yy[np.newaxis,...],zz[np.newaxis,:]],0).astype('float32')

        pad = []
        pad.append([0,0])
        for i in range(3):
            leftpad = max(0,-start[i])
            rightpad = max(0,start[i]+crop_size[i]-imgs.shape[i+1])
            pad.append([leftpad,rightpad])
        crop = imgs[:,
            max(start[0],0):min(start[0] + crop_size[0],imgs.shape[1]),
            max(start[1],0):min(start[1] + crop_size[1],imgs.shape[2]),
            max(start[2],0):min(start[2] + crop_size[2],imgs.shape[3])]
        crop = np.pad(crop,pad,'constant',constant_values =self.pad_value)
        for i in range(3):
            tem = [0, 2, 4]
            j = tem[i]
            target[j] = target[j] - start[i]
            target[j + 1] = target[j + 1] - start[i]
            # target[i] = target[i] - start[i]
        for i in range(len(bboxes)):
            for j in range(3):
                tem = [0, 2, 4]
                k = tem[j]
                bboxes[i][k] = bboxes[i][k] - start[j]
                bboxes[i][k + 1] = bboxes[i][k + 1] - start[j]
                # bboxes[i][j] = bboxes[i][j] - start[j]

        if isScale:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                crop = zoom(crop,[1,scale,scale,scale],order=1)
            newpad = self.crop_size[0]-crop.shape[1:][0]
            if newpad<0:
                crop = crop[:,:-newpad,:-newpad,:-newpad]
            elif newpad>0:
                pad2 = [[0,0],[0,newpad],[0,newpad],[0,newpad]]
                crop = np.pad(crop, pad2, 'constant', constant_values=self.pad_value)
            for i in range(6):
                target[i] = target[i]*scale
            for i in range(len(bboxes)):
                for j in range(6):
                    bboxes[i][j] = bboxes[i][j]*scale
        return crop, target, bboxes, coord




import numpy as np
import torch
import utils.util as util
import options.options as option
from models import create_model
import random
import nrrd
import os



def main(path_to_data, path_to_save_vol):
    need_HR = False
    has_mask = False
    data_merged = True
    ToTensor = util.ImgToTensor()

    # set-up dataset and label for conditional information
    uid = path_to_data.split('/')[-1]
    data, header = nrrd.read(path_to_data)
    data = data.transpose((2, 1, 0))
    vol_in = torch.from_numpy(np.expand_dims(data, axis=0))
    vol_in = torch.unsqueeze(ToTensor(vol_in), axis=0)
    class_label = {'d10':0, 'd25':1, 'd100': 2, 'k1':0, 'k2':1, 'k3':2, 'st1':0, 'st0.6':1, 'st2':2}
    if not data_merged:
        kernel_lbl, dose_lbl, st_lbl = self.opt['dataroot_LR'].split('/')[-1].split('_')
        kernel_lbl_to_assign, dose_lbl_to_assign, st_lbl_to_assign = util.cvt_int(kernel_lbl),\
                                                                                util.cvt_int(dose_lbl), util.cvt_int(st_lbl)
        kernel_lbl = torch.from_numpy(np.expand_dims(class_label[kernel_lbl_to_assign], axis=0))
        dose_lbl = torch.from_numpy(np.expand_dims(class_label[dose_lbl_to_assign], axis=0))
    else:
        kernel_lbl = torch.from_numpy(np.expand_dims(class_label[uid.split('_')[1:][0]], axis=0))
        dose_lbl = torch.from_numpy(np.expand_dims(class_label[uid.split('_')[1:][1]], axis=0))
    data = {'LR':vol_in, 'uid': uid, 'kernel': kernel_lbl, 'dose': dose_lbl}

    # set-up config
    opt = {
            'model': 'srgan',
            'scale': 1.0,
            'gpu_ids': [0],
            'precision': 'fp16',
            'result_format': 'nrrd',
            'need_label': True,
            "datasets": {"val": {"need_voxels": {"tile_x_y": True, "tile_size": 64},"slice_size": 32, "overlap_slice_size": 4, "LR_slice_size": 32, "LR_size": 64}},
            'path':{'pretrain_model_G': '../experiments/train_cSNGAN-AC-EQMergedData_tileStitch-32x64x64_Ref-k2d100/models/latest_G.pth'},
            'network_G':{'which_model_G': 'sr_resnet', 'norm_type': None, 'nf': 64, 'nb': 8, 'in_nc': 1, 'out_nc': 1,\
            "use_attention": False, 'need_embed':{'kernel_class': 3, 'dose_class': 3}, 'scale': 1.0},
            'is_train': False
        }
    opt = option.dict_to_nonedict(opt)
    torch.backends.cudnn.benchmark = True

    # create model and convert to fp16
    model = create_model(opt)
    model.half()

    if opt["precision"] == 'int8':
        model.prepare_quant(data, not_loader=True)
    model.feed_test_data(data, need_HR=need_HR)
    model.test(data)
    visuals = model.get_current_visuals(data, maskOn=has_mask, need_HR=need_HR)
    # save normalized visual
    try:    
        save_path = os.path.join(path_to_save_vol, 'norm_'+uid)
        sr_vol = util.tensor2img(visuals['SR'], out_type=np.uint16)
        nrrd.write(save_path, sr_vol, header, index_order='C')
        print('====== Volume Saved ======')
    except Exception as e:
        print(e)
    print('Exiting ...')
    exit()


if __name__ == '__main__':
    path_to_data = '/datasets/data_st1.0_merged/0a3e8ff1c12efe170b9b69c5f792772f_k1_d10_st1.nrrd'
    path_to_save_vol = '/datasets/reference'
    main(path_to_data, path_to_save_vol)


'''
Training code for MRBrainS18 datasets segmentation
Written by Whalechen
'''

from setting import parse_opts 
from datasets.brains18 import BrainS18Dataset 
from model import generate_model
import torch
import numpy as np
from torch import nn
from torch import optim
from torch.optim import lr_scheduler
from torch.utils.data import DataLoader
import time
from utils.logger import log
from scipy import ndimage
import os

def train(data_loader, model, optimizer, scheduler, total_epochs, save_interval, save_folder, sets):
    # settings
    batches_per_epoch = len(data_loader)
    log.info('{} epochs in total, {} batches per epoch'.format(total_epochs, batches_per_epoch))
    loss_seg = nn.CrossEntropyLoss(ignore_index=-1)

    print("Current setting is:")
    print(sets)
    print("\n\n")     
    if not sets.no_cuda:
        loss_seg = loss_seg.cuda()
        
    model.train()
    train_time_sp = time.time()
    for epoch in range(total_epochs):
        log.info('Start epoch {}'.format(epoch))
        
        scheduler.step()
        log.info('lr = {}'.format(scheduler.get_lr()))
        
        for batch_id, batch_data in enumerate(data_loader):
            # getting data batch
            batch_id_sp = epoch * batches_per_epoch
            volumes, label_masks = batch_data

            if not sets.no_cuda: 
                volumes = volumes.cuda()

            optimizer.zero_grad()
            out_masks = model(volumes)
            # resize label
            [n, _, d, h, w] = out_masks.shape
            new_label_masks = np.zeros([n, d, h, w])
            for label_id in range(n):
                label_mask = label_masks[label_id]
                [ori_c, ori_d, ori_h, ori_w] = label_mask.shape 
                label_mask = np.reshape(label_mask, [ori_d, ori_h, ori_w])
                scale = [d*1.0/ori_d, h*1.0/ori_h, w*1.0/ori_w]
                label_mask = ndimage.interpolation.zoom(label_mask, scale, order=0)
                new_label_masks[label_id] = label_mask

            new_label_masks = torch.tensor(new_label_masks).to(torch.int64)
            if not sets.no_cuda:
                new_label_masks = new_label_masks.cuda()

            # calculating loss
            loss_value_seg = loss_seg(out_masks, new_label_masks)
            loss = loss_value_seg
            loss.backward()                
            optimizer.step()

            avg_batch_time = (time.time() - train_time_sp) / (1 + batch_id_sp)
            log.info(
                    'Batch: {}-{} ({}), loss = {:.3f}, loss_seg = {:.3f}, avg_batch_time = {:.3f}'\
                    .format(epoch, batch_id, batch_id_sp, loss.item(), loss_value_seg.item(), avg_batch_time))
          
            if not sets.ci_test:
                # save model
                if batch_id == 0 and batch_id_sp != 0 and batch_id_sp % save_interval == 0:
                #if batch_id_sp != 0 and batch_id_sp % save_interval == 0:
                    model_save_path = '{}_epoch_{}_batch_{}.pth.tar'.format(save_folder, epoch, batch_id)
                    model_save_dir = os.path.dirname(model_save_path)
                    if not os.path.exists(model_save_dir):
                        os.makedirs(model_save_dir)
                    
                    log.info('Save checkpoints: epoch = {}, batch_id = {}'.format(epoch, batch_id)) 
                    torch.save({
                                'ecpoch': epoch,
                                'batch_id': batch_id,
                                'state_dict': model.state_dict(),
                                'optimizer': optimizer.state_dict()},
                                model_save_path)
                            
    print('Finished training')            
    if sets.ci_test:
        exit()


if __name__ == '__main__':
    # settting
    sets = parse_opts()   
    if sets.ci_test:
        sets.img_list = './toy_data/test_ci.txt' 
        sets.n_epochs = 1
        sets.no_cuda = True
        sets.data_root = './toy_data'
        sets.pretrain_path = ''
        sets.num_workers = 0
        sets.model_depth = 10
        sets.resnet_shortcut = 'A'
        sets.input_D = 14
        sets.input_H = 28
        sets.input_W = 28
       
    # getting model
    torch.manual_seed(sets.manual_seed)
    model, parameters = generate_model(sets) 
    print (model)
    # optimizer
    if sets.ci_test:
        params = [{'params': parameters, 'lr': sets.learning_rate}]
    else:
        params = [
                { 'params': parameters['base_parameters'], 'lr': sets.learning_rate }, 
                { 'params': parameters['new_parameters'], 'lr': sets.learning_rate*100 }
                ]
    optimizer = torch.optim.SGD(params, momentum=0.9, weight_decay=1e-3)   
    scheduler = optim.lr_scheduler.ExponentialLR(optimizer, gamma=0.99)
    
    # train from resume
    if sets.resume_path:
        if os.path.isfile(sets.resume_path):
            print("=> loading checkpoint '{}'".format(sets.resume_path))
            checkpoint = torch.load(sets.resume_path)
            model.load_state_dict(checkpoint['state_dict'])
            optimizer.load_state_dict(checkpoint['optimizer'])
            print("=> loaded checkpoint '{}' (epoch {})"
              .format(sets.resume_path, checkpoint['epoch']))

    # getting data
    sets.phase = 'train'
    if sets.no_cuda:
        sets.pin_memory = False
    else:
        sets.pin_memory = True    
    training_dataset = BrainS18Dataset(sets.data_root, sets.img_list, sets)
    data_loader = DataLoader(training_dataset, batch_size=sets.batch_size, shuffle=True, num_workers=sets.num_workers, pin_memory=sets.pin_memory)

    # training
    train(data_loader, model, optimizer, scheduler, total_epochs=sets.n_epochs, save_interval=sets.save_intervals, save_folder=sets.save_folder, sets=sets) 


import numpy as np
import torch
from torch.utils.data import Dataset
import os, imageio
from scipy.ndimage import zoom
import warnings
from scipy.ndimage.interpolation import rotate
import scipy
import math
import time
import nrrd
import pandas as pd
import SimpleITK as sitk
import h5py
import json
from utils.util import normalize, truncate_HU_uint8


class BboxReader(Dataset):
    def __init__(self, data_dir, set_name, cfg, mode='train', resample_vol=False, segment_lung=False, add_cnd=False):
        print('Inside BboxReader Dataset')
        print('-'*40)
        self.mode = mode
        print('mode:', self.mode)
        self.cfg = cfg
        self.r_rand = cfg['r_rand_crop']
        print('r_rand:', self.r_rand)
        self.augtype = cfg['augtype']
        print('augtype:', self.augtype)
        self.pad_value = cfg['pad_value']
        print('pad_value:', self.pad_value)
        self.data_dir = data_dir
        print('data_dir:', self.data_dir)
        self.stride = cfg['stride']
        print('stride:', self.stride)
        if segment_lung:
            print('Will segment lung')
            self.path_to_mask = segment_lung
        else:
            print('Will not segment lung')
            self.path_to_mask = None

        self.blacklist = cfg['blacklist']
        # print('blacklist:', self.blacklist)
        self.set_name = set_name
        self.need_resample = resample_vol
        self.add_cnd = add_cnd
        # print('set_name:', set_name)
        sizelim = 0
        sizelim2 = 10
        sizelim3 = 20

        labels = []
        with open(self.set_name, "r") as f:
            self.filenames = f.read().splitlines()

        if mode != 'test':
            self.filenames = [f for f in self.filenames if (f not in self.blacklist)]
        if self.mode in ['train', 'val']:
            csv_dir = cfg['train_anno']
        else:
            csv_dir = cfg['test_anno']

        # print('csv_dir:', csv_dir)
        annos_all = pd.read_csv(csv_dir)
        # print(annos_all.head())
        # print('-'*40)
        # collecting annotations
        for fn in self.filenames:
            annos = annos_all[annos_all['pid'] == fn]            
            temp_annos = []
            if len(annos) > 0:
                for index in range(len(annos)):
                    anno = annos.iloc[index]
                    temp_annos.append([anno['zmin'], anno['zmax'], anno['ymin'], anno['ymax'], anno['xmin'], anno['xmax']])
            l = np.array(temp_annos)
            if np.all(l==0):
                l=np.array([])
            labels.append(l)

        self.sample_bboxes = labels
        # print('all labels:', self.sample_bboxes)
        if self.mode in ['train', 'val']:
            self.bboxes = []
            for i, l in enumerate(labels):
                if len(l) > 0 :
                    for t in l:
                        self.bboxes.append([np.concatenate([[i],t])])
            self.bboxes = np.concatenate(self.bboxes, axis=0).astype(np.float32)
            # print('self.bboxes:', self.bboxes)
        self.crop = Crop(cfg)

    def __getitem__(self, idx):
        # print('Getting index:', idx)
        t = time.time()
        np.random.seed(int(str(t % 1)[2:7]))  # seed according to time
        is_random_img  = False
        if self.mode in ['train', 'val']:
            if idx >= len(self.bboxes):
                is_random_crop = True
                idx = idx % len(self.bboxes)
                is_random_img = np.random.randint(2)
            else:
                is_random_crop = False
        else:
            is_random_crop = False
        print('is_random_crop', False)

        if self.mode in ['train', 'val']:
            print('Entering in train or val')
            if not is_random_img:
                bbox = self.bboxes[idx]
                filename = self.filenames[int(bbox[0])]
                imgs = np.load(os.path.join(self.data_dir, '%s_zoom.npy' % (filename)))

                # lines = os.listdir(os.path.join(self.data_dir.replace('full', 'img'), filename))
                # lines = sorted(lines)
                # slice_files = [os.path.join(self.data_dir.replace('full', 'img'), filename, s) for s in lines]
                # slices = [imageio.imread(s) for s in slice_files]
                # imgs = np.array(slices)
                # imgs = imgs[np.newaxis, ...]

                bboxes = self.sample_bboxes[int(bbox[0])]
                isScale = self.augtype['scale'] and (self.mode=='train')
                sample, target, bboxes, coord = self.crop(imgs, bbox[1:], bboxes,isScale,is_random_crop)
                if self.mode == 'train' and not is_random_crop:
                     sample, target, bboxes = augment(sample, target, bboxes, do_flip = self.augtype['flip'], 
                                                             do_rotate=self.augtype['rotate'], do_swap = self.augtype['swap'])
            else:
                randimid = np.random.randint(len(self.filenames))
                filename = self.filenames[randimid]
                imgs = np.load(os.path.join(self.data_dir, '%s_zoom.npy' % (filename)))

                # lines = os.listdir(os.path.join(self.data_dir.replace('full', 'img'), filename))
                # lines = sorted(lines)
                # slice_files = [os.path.join(self.data_dir.replace('full', 'img'), filename, s) for s in lines]
                # slices = [imageio.imread(s) for s in slice_files]
                # imgs = np.array(slices)
                # imgs = imgs[np.newaxis, ...]

                bboxes = self.sample_bboxes[randimid]
                isScale = self.augtype['scale'] and (self.mode=='train')
                sample, target, bboxes, coord = self.crop(imgs, [], bboxes,isScale=False,isRand=True)

            if sample.shape[1] != self.cfg['crop_size'][0] or sample.shape[2] != \
                self.cfg['crop_size'][1] or sample.shape[3] != self.cfg['crop_size'][2]:
                print(filename, sample.shape)

            sample = (sample.astype(np.float32)-128)/128
            bboxes = fillter_box(bboxes, self.cfg['crop_size'])
            bboxes = corner_form_to_center_form(bboxes, self.cfg['bbox_border'])
            bboxes = np.array(bboxes)
            truth_labels = bboxes[:, -1]
            truth_bboxes = bboxes[:, :-1]

            return [torch.from_numpy(sample).float(), truth_bboxes, truth_labels]

        if self.mode in ['eval']:
            # print('Mode is in eval')
            filename = self.filenames[idx]
            if self.add_cnd:
                filename_to_read = filename+'_{}'.format(self.add_cnd)
            else:
                filename_to_read = filename

            print('Reading file:', filename_to_read)
            # read file
            if os.path.isfile(os.path.join(self.data_dir, filename_to_read+'.h5')):
                with h5py.File(os.path.join(self.data_dir, filename_to_read+'.h5'), 'r') as file:
                    file_data = file['data'][:,:,:]
                    header = read_config(filename)
            elif os.path.isfile(os.path.join(self.data_dir, filename_to_read+'.nrrd')):
                file_data, header = nrrd.read(os.path.join(self.data_dir, filename_to_read+'.nrrd'))
                file_data = file_data.transpose((2, 1, 0))
            else:
                raise ValueError('File not found with either .h5 or .nrrd extension!...Exiting')

            # numpyImage = np.clip(file_data, 0, 1500.).astype(np.float32)
            if self.need_resample:
                spacing = [header['spacings'][2], header['spacings'][0], header['spacings'][1]]
                file_data, _ = resample_vol(file_data, orig_spacing=spacing)
                # print('resample shape:', file_data.shape)
                # print('min-max:', file_data.min(), file_data.max())

            numpyImage = bound_image(file_data).astype(np.float32)
            # print('after bound:', numpyImage.min(), numpyImage.max())
            numpyImage /= 1500.0
            numpyImage *= 255.0
            numpyImage = numpyImage.astype(np.uint8)
            print(numpyImage.shape, numpyImage.min(), numpyImage.max())
            if self.path_to_mask:
                with h5py.File(os.path.join(self.path_to_mask, filename+'.h5'), 'r') as mask_file:
                    mask_data = mask_file['data'][:,:,:]
                coords = np.where(mask_data == 0)
                numpyImage[coords] = 0

            image = pad2factor(numpyImage)
            image = np.expand_dims(image, 0)
            bboxes = self.sample_bboxes[idx]

            bboxes = corner_form_to_center_form(bboxes, self.cfg['bbox_border'])
            bboxes = np.array(bboxes)
            truth_labels = bboxes[:, -1]
            truth_bboxes = bboxes[:, :-1]

            input = normalize(image.astype(np.float32))
            # input = (image.astype(np.float32) - 128.) / 128.
            return [torch.from_numpy(input).float(), truth_bboxes, truth_labels]


    def __len__(self):
        if self.mode == 'train':
            return int(len(self.bboxes) / (1-self.r_rand))//self.cfg['batch_size']*self.cfg['batch_size']
        elif self.mode =='val':
            return len(self.bboxes)//self.cfg['batch_size']*self.cfg['batch_size']
        else:
            return len(self.filenames)


"""
Used only for '.h5' files that have header information in json file
"""
def read_config(json_uid):
    config_path = os.path.join('/datasets/reference/k2_d100_st1/', json_uid+'.json')
    with open(config_path, 'r') as f:
        config = json.load(f)
    meta_data = {}
    meta_data['Spacing'] = [float(i) for i in config['Spacing'].split()]
    meta_data['Orientation'] = [float(i) for i in config['Orientation'].split()] 
    meta_data['Origin'] = [float(i) for i in config['Origin'].split()]
    header = {'units': ['mm', 'mm', 'mm'], 'spacings': meta_data['Spacing']} 
    return header

def bound_image(img, lower_bound=0, upper_bound=1500, shift_pixels=0):
    image_data = img.astype('float32')
    image_data[image_data < lower_bound] = lower_bound # Sets all values less than -1000 as -1000
    image_data[image_data > upper_bound] = upper_bound # Sets all values greater than 500 as 500
    shifted_pixels = image_data + shift_pixels
    shifted_pixels = shifted_pixels.astype(np.int16)
    return shifted_pixels

def resample_vol(image, orig_spacing, new_spacing=[1,1,1]):
    # Determine current pixel spacing
    spacing = np.array(orig_spacing)
    resize_factor = spacing / new_spacing
    new_real_shape = image.shape * resize_factor
    new_shape = np.round(new_real_shape)
    real_resize_factor = new_shape / image.shape
    new_spacing = spacing / real_resize_factor
    image = scipy.ndimage.interpolation.zoom(image, real_resize_factor)
    return image, new_spacing


def corner_form_to_center_form(boxes, border):
    # print('converting bbox for corner to center')
    # print('bboxes are:', boxes)
    # print('border:', border)
    bboxes = []
    for box in boxes:
        """
        print('[({}+{})/2., ({}+{})/2., ({}+{})/2., ({}-{}+1+{}), ({}-{}+1+{}), ({}-{}+1+{})]'.format(box[0], box[1], box[2], box[3],
                                                                                        box[4], box[5], box[1], box[0], border,
                                                                                        box[3], box[2], border,
                                                                                        box[5], box[4], border))
        """
        bboxes.append([(box[0] + box[1]) / 2.,
                       (box[2] + box[3]) / 2.,
                       (box[4] + box[5]) / 2.,
                       box[1] - box[0] + 1 + border,
                       box[3] - box[2] + 1 + border,
                       box[5] - box[4] + 1 + border,
                       1])

    # print('converted boxes:', bboxes)  
    return bboxes

def pad2factor(image, factor=32, pad_value=0):
    # print('Inside pad2factor')
    # print('-'*40)
    depth, height, width = image.shape
    # print('depth, height, width:', depth, height, width)
    d = int(math.ceil(depth / float(factor))) * factor
    # print("d = int(math.ceil({}/{})) * {} = {}".format(depth, factor, factor, d))
    h = int(math.ceil(height / float(factor))) * factor
    # print("h = int(math.ceil({}/{})) * {} = {}".format(height, factor, factor, h))
    w = int(math.ceil(width / float(factor))) * factor
    # print("w = int(math.ceil({}/{})) * {} = {}".format(width, factor, factor, w))

    pad = []
    pad.append([0, d - depth])
    pad.append([0, h - height])
    pad.append([0, w - width])
    # print('adding pad: [{}-{}, {}-{}, {}-{}] = [{}, {}, {}]'.format(d, depth, h, height, w, width, d-depth, h-height, w-width))

    image = np.pad(image, pad, 'constant', constant_values=pad_value)
    return image


def fillter_box(bboxes, size):
    res = []
    for box in bboxes:
        if np.all(box[:6] > 0) and np.all(box[:6]  < size[0]):
            res.append(box)
    return np.array(res)

def augment(sample, target, bboxes, do_flip = True, do_rotate=True, do_swap = True):
    #  angle1 = np.random.rand()*180
    if do_rotate:
        validrot = False
        counter = 0
        while not validrot:
            newtarget = np.copy(target)
            angle1 = np.random.rand()*180
            size = np.array(sample.shape[2:4]).astype('float')
            rotmat = np.array([[np.cos(angle1/180*np.pi),-np.sin(angle1/180*np.pi)],[np.sin(angle1/180*np.pi),np.cos(angle1/180*np.pi)]])
            newtarget[1:3] = np.dot(rotmat,target[1:3]-size/2)+size/2
            if np.all(newtarget[:3]>target[3]) and np.all(newtarget[:3]< np.array(sample.shape[1:4])-newtarget[3]):
                validrot = True
                target = newtarget
                sample = rotate(sample,angle1,axes=(2,3),reshape=False)
                for box in bboxes:
                    box[1:3] = np.dot(rotmat,box[1:3]-size/2)+size/2
            else:
                counter += 1
                if counter ==3:
                    break
    if do_swap:
        if sample.shape[1]==sample.shape[2] and sample.shape[1]==sample.shape[3]:
            axisorder = np.random.permutation(3)
            sample = np.transpose(sample,np.concatenate([[0],axisorder+1]))
            target[:3] = target[:3][axisorder]
            bboxes[:,:3] = bboxes[:,:3][:,axisorder]

    if do_flip:
        # flipid = np.array([np.random.randint(2),np.random.randint(2),np.random.randint(2)])*2-1
        flipid = np.array([1,np.random.randint(2),np.random.randint(2)])*2-1
        sample = np.ascontiguousarray(sample[:,::flipid[0],::flipid[1],::flipid[2]])
        # for ax in range(3):
        #     if flipid[ax]==-1:
        #         target[ax] = np.array(sample.shape[ax+1])-target[ax]
        #         bboxes[:,ax]= np.array(sample.shape[ax+1])-bboxes[:,ax]
        for i in range(3):
            if flipid[i]==-1:
                tem = [0, 2, 4]
                ax = tem[i]
                target[ax] = np.array(sample.shape[i+1])-target[ax]
                bboxes[:,ax]= np.array(sample.shape[i+1])-bboxes[:,ax]
                target[ax+1] = np.array(sample.shape[i + 1]) - target[ax+1]
                bboxes[:, ax+1] = np.array(sample.shape[i + 1]) - bboxes[:, ax+1]
    return sample, target, bboxes

class Crop(object):
    # print('Inside Crop Class')
    # print('-'*40)
    def __init__(self, config):
        self.crop_size = config['crop_size']
        self.bound_size = config['bound_size']
        self.stride = config['stride']
        self.pad_value = config['pad_value']
    def __call__(self, imgs, target, bboxes,isScale=False,isRand=False):
        if isScale:
            radiusLim = [8.,120.]
            scaleLim = [0.75,1.25]
            scaleRange = [np.min([np.max([(radiusLim[0]/target[3]),scaleLim[0]]),1])
                         ,np.max([np.min([(radiusLim[1]/target[3]),scaleLim[1]]),1])]
            scale = np.random.rand()*(scaleRange[1]-scaleRange[0])+scaleRange[0]
            crop_size = (np.array(self.crop_size).astype('float')/scale).astype('int')
        else:
            crop_size=self.crop_size
        bound_size = self.bound_size
        target = np.copy(target)
        bboxes = np.copy(bboxes)

        start = []
        for i in range(3):
            # start.append(int(target[i] - crop_size[i] / 2))
            if not isRand:
                tem = [0,2,4]
                j = tem[i]
                # r = target[3] / 2
                # s = np.floor(target[i] - r)+ 1 - bound_size
                # e = np.ceil (target[i] + r)+ 1 + bound_size - crop_size[i]
                s = target[j] - bound_size
                e = target[j+1] + bound_size - crop_size[i]
            else:
                s = np.max([imgs.shape[i+1]-crop_size[i]/2,imgs.shape[i+1]/2+bound_size])
                e = np.min([crop_size[i]/2, imgs.shape[i+1]/2-bound_size])
                target = np.array([np.nan,np.nan,np.nan,np.nan])
            if s>e:
                start.append(np.random.randint(e,s))#!
            else:
                start.append(int(target[i])-int(crop_size[i]/2)+np.random.randint(-bound_size/2,bound_size/2))


        normstart = np.array(start).astype('float32')/np.array(imgs.shape[1:])-0.5
        normsize = np.array(crop_size).astype('float32')/np.array(imgs.shape[1:])
        xx,yy,zz = np.meshgrid(np.linspace(normstart[0],normstart[0]+normsize[0],int(self.crop_size[0]/self.stride)),
                           np.linspace(normstart[1],normstart[1]+normsize[1],int(self.crop_size[1]/self.stride)),
                           np.linspace(normstart[2],normstart[2]+normsize[2],int(self.crop_size[2]/self.stride)),indexing ='ij')
        coord = np.concatenate([xx[np.newaxis,...], yy[np.newaxis,...],zz[np.newaxis,:]],0).astype('float32')

        pad = []
        pad.append([0,0])
        for i in range(3):
            leftpad = max(0,-start[i])
            rightpad = max(0,start[i]+crop_size[i]-imgs.shape[i+1])
            pad.append([leftpad,rightpad])
        crop = imgs[:,
            max(start[0],0):min(start[0] + crop_size[0],imgs.shape[1]),
            max(start[1],0):min(start[1] + crop_size[1],imgs.shape[2]),
            max(start[2],0):min(start[2] + crop_size[2],imgs.shape[3])]
        crop = np.pad(crop,pad,'constant',constant_values =self.pad_value)
        for i in range(3):
            tem = [0, 2, 4]
            j = tem[i]
            target[j] = target[j] - start[i]
            target[j + 1] = target[j + 1] - start[i]
            # target[i] = target[i] - start[i]
        for i in range(len(bboxes)):
            for j in range(3):
                tem = [0, 2, 4]
                k = tem[j]
                bboxes[i][k] = bboxes[i][k] - start[j]
                bboxes[i][k + 1] = bboxes[i][k + 1] - start[j]
                # bboxes[i][j] = bboxes[i][j] - start[j]

        if isScale:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                crop = zoom(crop,[1,scale,scale,scale],order=1)
            newpad = self.crop_size[0]-crop.shape[1:][0]
            if newpad<0:
                crop = crop[:,:-newpad,:-newpad,:-newpad]
            elif newpad>0:
                pad2 = [[0,0],[0,newpad],[0,newpad],[0,newpad]]
                crop = np.pad(crop, pad2, 'constant', constant_values=self.pad_value)
            for i in range(6):
                target[i] = target[i]*scale
            for i in range(len(bboxes)):
                for j in range(6):
                    bboxes[i][j] = bboxes[i][j]*scale
        return crop, target, bboxes, coord




import torch
import torch.nn as nn
import torch.nn.functional as F
from collections import OrderedDict


class double_conv(nn.Module):
    '''(conv => BN => ReLU) * 2'''
    def __init__(self, in_ch, out_ch):
        super(double_conv, self).__init__()
        self.conv = nn.Sequential(
            nn.Conv3d(in_ch, out_ch, kernel_size=3, stride=1, padding=1),
            nn.BatchNorm3d(out_ch),
            nn.ReLU(inplace=True),
            nn.Conv3d(out_ch, out_ch, kernel_size=3, stride=1, padding=1),
            nn.BatchNorm3d(out_ch),
            nn.ReLU(inplace=True)
        )

    def forward(self, x):
        x = self.conv(x)
        return x


class inconv(nn.Module):
    def __init__(self, in_ch, out_ch):
        super(inconv, self).__init__()
        self.conv = double_conv(in_ch, out_ch)

    def forward(self, x):
        x = self.conv(x)
        return x


class down(nn.Module):
    def __init__(self, in_ch, out_ch):
        super(down, self).__init__()
        self.mpconv = nn.Sequential(
            nn.MaxPool3d(2),
            double_conv(in_ch, out_ch)
        )

    def forward(self, x):
        x = self.mpconv(x)
        return x


class up(nn.Module):
    def __init__(self, in_ch, out_ch, bilinear=True):
        super(up, self).__init__()
        if bilinear:
            # self.up = nn.UpsamplingBilinear2d(scale_factor=2)
            self.up = nn.Upsample(scale_factor=2)
        else:
            self.up = nn.ConvTranspose3d(in_ch//2, in_ch//2, 2, stride=2)

        self.conv = double_conv(in_ch, out_ch)

    def forward(self, x1, x2):
        x1 = self.up(x1)
        diffX = x1.size()[2] - x2.size()[2]
        diffY = x1.size()[3] - x2.size()[3]
        x2 = F.pad(x2, (diffX // 2, int(diffX / 2),
                        diffY // 2, int(diffY / 2)))
        x = torch.cat([x2, x1], dim=1)
        x = self.conv(x)
        return x
        

class outconv(nn.Module):
    def __init__(self, in_ch, out_ch):
        super(outconv, self).__init__()
        self.conv = nn.Conv3d(in_ch, out_ch, 1)

    def forward(self, x):
        x = self.conv(x)
        return x


class UNet(nn.Module):
    def __init__(self, in_nc, out_nc, nf):
        super(UNet, self).__init__()
        self.inc = inconv(in_nc, nf)
        self.down1 = down(nf, 128)
        self.down2 = down(128, 256)
        self.down3 = down(256, 512)
        self.down4 = down(512, 512)
        self.up1 = up(1024, 256)
        self.up2 = up(512, 128)
        self.up3 = up(256, nf)
        self.up4 = up(128, nf)
        self.outc = outconv(nf, out_nc)
        self.Dropout = nn.Dropout(0.2)

    def forward(self, x):
        # print('-'*40)
        # print('incoming x shape:', x.shape)
        x1 = self.inc(x)
        # print('shape after first inc:', x1.shape)
        x2 = self.down1(x1)
        # print('shape after first down1:', x2.shape)
        x3 = self.down2(x2)
        # print('shape after first down2:', x3.shape)
        x3 = self.Dropout(x3)
        # print('shape after first dropout:', x3.shape)
        x4 = self.down3(x3)
        # print('shape after first down3:', x4.shape)
        x5 = self.down4(x4)
        # print('shape after first down4:', x5.shape)
        x = self.up1(x5, x4)
        # print('shape after first up1:', x.shape)
        x = self.up2(x, x3)
        # print('shape after first up2:', x.shape)
        x = self.up3(x, x2)
        # print('shape after first up3:', x.shape)
        x = self.up4(x, x1)
        # print('shape after first up4:', x.shape)
        x = self.outc(x)
        # print('final output:', x.shape)
        return x


import logging
logger = logging.getLogger('base')
from collections import OrderedDict
import os
import lung_analysis_pipeline.utils.util as util
from ..non_DL import bm3d_ht as bm3d
import pkg_resources

"""
Create a normalizatio model given opt 
"""
def create_model(opt):
    model = opt['normalization']['model']
    ## convert opt to normalization module option format 
    if model == 'bm3d':
        return util.PythonFunctionWrapper(bm3d.calculate_scan_bm3d)
    elif model == 'wgan': 
        model = 'srgan'
        which_model_G = 'vanilla'
        # pretrain_model_G = "Normalization_Pipeline/weights/WGAN/latest_G.pth"
    elif model == 'sngan': 
        model = 'srgan'
        which_model_G = 'sr_resnet'
        # pretrain_model_G = "Normalization_Pipeline/weights/SNGAN/latest_G.pth"
    elif model == 'sr_resnet': 
        model = 'sr'
        which_model_G = 'sr_resnet'
        # pretrain_model_G = "Normalization_Pipeline/weights/SRResNet/latest_G.pth"
    elif model == 'rrdb': 
        model = 'sr'
        which_model_G = 'RRDB'
        # pretrain_model_G = "Normalization_Pipeline/weights/RRDB/latest_G.pth"
    else:
        raise NotImplementedError('Model [{:s}] not implemented.'.format(opt['normalization']['model']))
    # convert relative path to path in pip package 
    # pretrain_model_G = pkg_resources.resource_filename('lung_analysis_pipeline', pretrain_model_G)
    pretrain_model_G = opt['normalization']['weights']

    scale = 1.0
    option = OrderedDict([('name', opt['normalization']['model'])])
    option['is_train'] = False
    option['model'] = model
    option['scale'] = scale
    option['gpu_ids'] = opt['gpu_ids']
    option['precision'] = 'fp16'
    option["result_format"] = opt['normalization']['output_type']
    option["need_label"] = False
    option["data_merged"] = True
    option['train'] = None

    # dataset 
    dataset = OrderedDict([('data_type', opt['dataset']['image_type'])])
    dataset['name'] = opt['dataset']['source'] 
    dataset['uids_path'] = opt['dataset']['uids_location']
    dataset['dataroot_HR'] = opt['dataset']['dataroot_HR']
    dataset['dataroot_LR'] = opt['dataset']['dataroot_LR']
    dataset['need_voxels'] = OrderedDict([('tile_x_y', True), ('tile_size', 64)])
    dataset['slice_size'] = 32 
    dataset['overlap_slice_size'] = 4
    dataset['LR_slice_size'] = 32
    dataset['phase'] = 'val'
    dataset['scale'] = scale
    dataset['LR_size'] = 64
    dataset['dataset_name'] = opt['dataset']['source']
    option['datasets'] = OrderedDict([('val', dataset)])
    
    results_root = opt['normalization']['output_location']
    option['path'] = OrderedDict([
        ('results_root', results_root), 
        ('log', results_root),
        ("pretrain_model_G", pretrain_model_G)
    ])

    for key, path in option['path'].items():
        if path and key in option['path']:
            option['path'][key] = os.path.expanduser(path)

    # network 
    option['network_G'] = OrderedDict([
        ('which_model_G', which_model_G),
        ('norm_type', None), 
        ('nf', 64), 
        ('nb', 8), 
        ('in_nc', 1), 
        ('out_nc', 1), 
        ('use_attention', False), 
        ('need_embed', False), 
        ('scale', scale), 
    ])

    option = util.dict_to_nonedict(option)

    ## instantiate only G
    if model == 'sr':
        from .SR_model import SRModel as M
    ## 'srgan' instantiates both G and D
    elif model == 'srgan':
        from .SRGAN_model import SRGANModel as M
    elif model == 'srgan_exp':
        from .SRGAN_exp import SRGANModel as M
    else:
        raise NotImplementedError('Model [{:s}] not implemented.'.format(model))
    
    m = M(option)
    m.half()
    # copy net and convert to fp16 if necessary
    logger.info('Normalization model [{:s}] is created.'.format(m.__class__.__name__))
    logger.info(util.dict2str(option))

    return m
